"""
Module for running fmadm and fmdump on Solaris

:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:platform:      solaris,illumos

.. versionadded:: 2016.3.0
"""

import logging

import salt.utils.decorators as decorators
import salt.utils.path
import salt.utils.platform
from salt.utils.odict import OrderedDict

log = logging.getLogger(__name__)

# Function aliases
__func_alias__ = {
    "list_records": "list",
}

# Define the module's virtual name
__virtualname__ = "fmadm"


@decorators.memoize
def _check_fmadm():
    """
    Looks to see if fmadm is present on the system
    """
    return salt.utils.path.which("fmadm")


def _check_fmdump():
    """
    Looks to see if fmdump is present on the system
    """
    return salt.utils.path.which("fmdump")


def __virtual__():
    """
    Provides fmadm only on Solaris
    """
    if salt.utils.platform.is_sunos() and _check_fmadm() and _check_fmdump():
        return __virtualname__
    return (
        False,
        "{} module can only be loaded on Solaris with the fault management installed".format(
            __virtualname__
        ),
    )


def _parse_fmdump(output):
    """
    Parses fmdump output
    """
    result = []
    output = output.split("\n")

    # extract header
    header = [field for field in output[0].lower().split(" ") if field]
    del output[0]

    # parse entries
    for entry in output:
        entry = [item for item in entry.split(" ") if item]
        entry = ["{} {} {}".format(entry[0], entry[1], entry[2])] + entry[3:]

        # prepare faults
        fault = OrderedDict()
        for field in header:
            fault[field] = entry[header.index(field)]

        result.append(fault)

    return result


def _parse_fmdump_verbose(output):
    """
    Parses fmdump verbose output
    """
    result = []
    output = output.split("\n")

    fault = []
    verbose_fault = {}
    for line in output:
        if line.startswith("TIME"):
            fault.append(line)
            if len(verbose_fault) > 0:
                result.append(verbose_fault)
                verbose_fault = {}
        elif len(fault) == 1:
            fault.append(line)
            verbose_fault = _parse_fmdump("\n".join(fault))[0]
            fault = []
        elif len(verbose_fault) > 0:
            if "details" not in verbose_fault:
                verbose_fault["details"] = ""
            if line.strip() == "":
                continue
            verbose_fault["details"] = "{}{}\n".format(verbose_fault["details"], line)
    if len(verbose_fault) > 0:
        result.append(verbose_fault)

    return result


def _parse_fmadm_config(output):
    """
    Parsbb fmdump/fmadm output
    """
    result = []
    output = output.split("\n")

    # extract header
    header = [field for field in output[0].lower().split(" ") if field]
    del output[0]

    # parse entries
    for entry in output:
        entry = [item for item in entry.split(" ") if item]
        entry = entry[0:3] + [" ".join(entry[3:])]

        # prepare component
        component = OrderedDict()
        for field in header:
            component[field] = entry[header.index(field)]

        result.append(component)

    # keying
    keyed_result = OrderedDict()
    for component in result:
        keyed_result[component["module"]] = component
        del keyed_result[component["module"]]["module"]

    result = keyed_result

    return result


def _fmadm_action_fmri(action, fmri):
    """
    Internal function for fmadm.repqired, fmadm.replaced, fmadm.flush
    """
    ret = {}
    fmadm = _check_fmadm()
    cmd = "{cmd} {action} {fmri}".format(cmd=fmadm, action=action, fmri=fmri)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    result = {}
    if retcode != 0:
        result["Error"] = res["stderr"]
    else:
        result = True

    return result


def _parse_fmadm_faulty(output):
    """
    Parse fmadm faulty output
    """

    def _merge_data(summary, fault):
        result = {}
        uuid = summary["event-id"]
        del summary["event-id"]

        result[uuid] = OrderedDict()
        result[uuid]["summary"] = summary
        result[uuid]["fault"] = fault
        return result

    result = {}
    summary = []
    summary_data = {}
    fault_data = {}
    data_key = None

    for line in output.split("\n"):
        # we hit a divider
        if line.startswith("-"):
            if summary and summary_data and fault_data:
                # we have data, store it and reset
                result.update(_merge_data(summary_data, fault_data))

                summary = []
                summary_data = {}
                fault_data = {}
                continue
            else:
                # we don't have all data, colelct more
                continue

        # if we do not have the header, store it
        if not summary:
            summary.append(line)
            continue

        # if we have the header but no data, store the data and parse it
        if summary and not summary_data:
            summary.append(line)
            summary_data = _parse_fmdump("\n".join(summary))[0]
            continue

        # if we have a header and data, assume the other lines are details
        if summary and summary_data:
            # if line starts with a whitespace and we already have a key, append
            if line.startswith(" ") and data_key:
                fault_data[data_key] = "{}\n{}".format(
                    fault_data[data_key], line.strip()
                )
            # we have a key : value line, parse it
            elif ":" in line:
                line = line.split(":")
                data_key = line[0].strip()
                fault_data[data_key] = ":".join(line[1:]).strip()
                # note: for some reason Chassis_id is lobbed ofter Platform, fix that here
                if data_key == "Platform":
                    fault_data["Chassis_id"] = (
                        fault_data[data_key][fault_data[data_key].index("Chassis_id") :]
                        .split(":")[-1]
                        .strip()
                    )
                    fault_data[data_key] = fault_data[data_key][
                        0 : fault_data[data_key].index("Chassis_id")
                    ].strip()

    # we have data, store it and reset
    result.update(_merge_data(summary_data, fault_data))

    return result


def list_records(after=None, before=None):
    """
    Display fault management logs

    after : string
        filter events after time, see man fmdump for format

    before : string
        filter events before time, see man fmdump for format

    CLI Example:

    .. code-block:: bash

        salt '*' fmadm.list
    """
    ret = {}
    fmdump = _check_fmdump()
    cmd = "{cmd}{after}{before}".format(
        cmd=fmdump,
        after=" -t {}".format(after) if after else "",
        before=" -T {}".format(before) if before else "",
    )
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    result = {}
    if retcode != 0:
        result["Error"] = "error executing fmdump"
    else:
        result = _parse_fmdump(res["stdout"])

    return result


def show(uuid):
    """
    Display log details

    uuid: string
        uuid of fault

    CLI Example:

    .. code-block:: bash

        salt '*' fmadm.show 11b4070f-4358-62fa-9e1e-998f485977e1
    """
    ret = {}
    fmdump = _check_fmdump()
    cmd = "{cmd} -u {uuid} -V".format(cmd=fmdump, uuid=uuid)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    result = {}
    if retcode != 0:
        result["Error"] = "error executing fmdump"
    else:
        result = _parse_fmdump_verbose(res["stdout"])

    return result


def config():
    """
    Display fault manager configuration

    CLI Example:

    .. code-block:: bash

        salt '*' fmadm.config
    """
    ret = {}
    fmadm = _check_fmadm()
    cmd = "{cmd} config".format(cmd=fmadm)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    result = {}
    if retcode != 0:
        result["Error"] = "error executing fmadm config"
    else:
        result = _parse_fmadm_config(res["stdout"])

    return result


def load(path):
    """
    Load specified fault manager module

    path: string
        path of fault manager module

    CLI Example:

    .. code-block:: bash

        salt '*' fmadm.load /module/path
    """
    ret = {}
    fmadm = _check_fmadm()
    cmd = "{cmd} load {path}".format(cmd=fmadm, path=path)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    result = {}
    if retcode != 0:
        result["Error"] = res["stderr"]
    else:
        result = True

    return result


def unload(module):
    """
    Unload specified fault manager module

    module: string
        module to unload

    CLI Example:

    .. code-block:: bash

        salt '*' fmadm.unload software-response
    """
    ret = {}
    fmadm = _check_fmadm()
    cmd = "{cmd} unload {module}".format(cmd=fmadm, module=module)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    result = {}
    if retcode != 0:
        result["Error"] = res["stderr"]
    else:
        result = True

    return result


def reset(module, serd=None):
    """
    Reset module or sub-component

    module: string
        module to unload
    serd : string
        serd sub module

    CLI Example:

    .. code-block:: bash

        salt '*' fmadm.reset software-response
    """
    ret = {}
    fmadm = _check_fmadm()
    cmd = "{cmd} reset {serd}{module}".format(
        cmd=fmadm, serd="-s {} ".format(serd) if serd else "", module=module
    )
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    result = {}
    if retcode != 0:
        result["Error"] = res["stderr"]
    else:
        result = True

    return result


def flush(fmri):
    """
    Flush cached state for resource

    fmri: string
        fmri

    CLI Example:

    .. code-block:: bash

        salt '*' fmadm.flush fmri
    """
    return _fmadm_action_fmri("flush", fmri)


def repaired(fmri):
    """
    Notify fault manager that resource has been repaired

    fmri: string
        fmri

    CLI Example:

    .. code-block:: bash

        salt '*' fmadm.repaired fmri
    """
    return _fmadm_action_fmri("repaired", fmri)


def replaced(fmri):
    """
    Notify fault manager that resource has been replaced

    fmri: string
        fmri

    CLI Example:

    .. code-block:: bash

        salt '*' fmadm.repaired fmri
    """
    return _fmadm_action_fmri("replaced", fmri)


def acquit(fmri):
    """
    Acquit resource or acquit case

    fmri: string
        fmri or uuid

    CLI Example:

    .. code-block:: bash

        salt '*' fmadm.acquit fmri | uuid
    """
    return _fmadm_action_fmri("acquit", fmri)


def faulty():
    """
    Display list of faulty resources

    CLI Example:

    .. code-block:: bash

        salt '*' fmadm.faulty
    """
    fmadm = _check_fmadm()
    cmd = "{cmd} faulty".format(
        cmd=fmadm,
    )
    res = __salt__["cmd.run_all"](cmd)
    result = {}
    if res["stdout"] == "":
        result = False
    else:
        result = _parse_fmadm_faulty(res["stdout"])

    return result


def healthy():
    """
    Return whether fmadm is reporting faults

    CLI Example:

    .. code-block:: bash

        salt '*' fmadm.healthy
    """
    return False if faulty() else True


# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import salt.modules.cassandra_mod as cassandra
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class CassandraTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.cassandra_mod
    """

    def setup_loader_modules(self):
        return {cassandra: {}}

    def test_compactionstats(self):
        """
        Test for Return compactionstats info
        """
        mock = MagicMock(return_value="A")
        with patch.object(cassandra, "_nodetool", mock):
            self.assertEqual(cassandra.compactionstats(), "A")

    def test_version(self):
        """
        Test for Return the cassandra version
        """
        mock = MagicMock(return_value="A")
        with patch.object(cassandra, "_nodetool", mock):
            self.assertEqual(cassandra.version(), "A")

    def test_netstats(self):
        """
        Test for Return netstats info
        """
        mock = MagicMock(return_value="A")
        with patch.object(cassandra, "_nodetool", mock):
            self.assertEqual(cassandra.netstats(), "A")

    def test_tpstats(self):
        """
        Test for Return tpstats info
        """
        mock = MagicMock(return_value="A")
        with patch.object(cassandra, "_nodetool", mock):
            self.assertEqual(cassandra.tpstats(), "A")

    def test_info(self):
        """
        Test for Return cassandra node info
        """
        mock = MagicMock(return_value="A")
        with patch.object(cassandra, "_nodetool", mock):
            self.assertEqual(cassandra.info(), "A")

    def test_ring(self):
        """
        Test for Return ring info
        """
        mock = MagicMock(return_value="A")
        with patch.object(cassandra, "_nodetool", mock):
            self.assertEqual(cassandra.ring(), "A")

    def test_keyspaces(self):
        """
        Test for Return existing keyspaces
        """
        mock_keyspaces = ["A", "B", "C", "D"]

        class MockSystemManager:
            def list_keyspaces(self):
                return mock_keyspaces

        mock_sys_mgr = MagicMock(return_value=MockSystemManager())

        with patch.object(cassandra, "_sys_mgr", mock_sys_mgr):
            self.assertEqual(cassandra.keyspaces(), mock_keyspaces)

    def test_column_families(self):
        """
        Test for Return existing column families for all keyspaces
        """
        mock_keyspaces = ["A", "B"]

        class MockSystemManager:
            def list_keyspaces(self):
                return mock_keyspaces

            def get_keyspace_column_families(self, keyspace):
                if keyspace == "A":
                    return {"a": "saltines", "b": "biscuits"}
                if keyspace == "B":
                    return {"c": "cheese", "d": "crackers"}

        mock_sys_mgr = MagicMock(return_value=MockSystemManager())

        with patch.object(cassandra, "_sys_mgr", mock_sys_mgr):
            self.assertEqual(cassandra.column_families("Z"), None)
            self.assertCountEqual(cassandra.column_families("A"), ["a", "b"])
            self.assertCountEqual(
                cassandra.column_families(), {"A": ["a", "b"], "B": ["c", "d"]}
            )

    def test_column_family_definition(self):
        """
        Test for Return a dictionary of column family definitions for the given
        keyspace/column_family
        """

        class MockSystemManager:
            def get_keyspace_column_families(self, keyspace):
                if keyspace == "A":
                    return {"a": object, "b": object}
                if keyspace == "B":
                    raise Exception

        mock_sys_mgr = MagicMock(return_value=MockSystemManager())

        with patch.object(cassandra, "_sys_mgr", mock_sys_mgr):
            self.assertEqual(cassandra.column_family_definition("A", "a"), vars(object))
            self.assertEqual(cassandra.column_family_definition("B", "a"), None)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import salt.states.gnomedesktop as gnomedesktop
from tests.support.unit import TestCase


class GnomedesktopTestCase(TestCase):
    """
    Test cases for salt.states.gnomedesktop
    """

    # 'wm_preferences' function tests: 1

    def test_wm_preferences(self):
        """
        Test to sets values in the org.gnome.desktop.wm.preferences schema
        """
        name = "salt"

        ret = {"name": name, "result": True, "comment": "", "changes": {}}

        self.assertDictEqual(gnomedesktop.wm_preferences(name), ret)

    # 'desktop_lockdown' function tests: 1

    def test_desktop_lockdown(self):
        """
        Test to sets values in the org.gnome.desktop.lockdown schema
        """
        name = "salt"

        ret = {"name": name, "result": True, "comment": "", "changes": {}}

        self.assertDictEqual(gnomedesktop.desktop_lockdown(name), ret)

    # 'desktop_interface' function tests: 1

    def test_desktop_interface(self):
        """
        Test to sets values in the org.gnome.desktop.interface schema
        """
        name = "salt"

        ret = {"name": name, "result": True, "comment": "", "changes": {}}

        self.assertDictEqual(gnomedesktop.desktop_interface(name), ret)



"""
    salt.log.handlers
    ~~~~~~~~~~~~~~~~~

    .. versionadded:: 0.17.0

    Custom logging handlers to be used in salt.
"""


import logging

from salt._logging.handlers import (
    FileHandler,
    QueueHandler,
    RotatingFileHandler,
    StreamHandler,
    SysLogHandler,
    TemporaryLoggingHandler,
    WatchedFileHandler,
)

# from salt.utils.versions import warn_until_date
# warn_until_date(
#    '20220101',
#    'Please stop using \'{name}\' and instead use \'salt._logging.handlers\'. '
#    '\'{name}\' will go away after {{date}}.'.format(
#        name=__name__
#    )
# )

NullHandler = logging.NullHandler



# pylint: skip-file
from __future__ import absolute_import, division, print_function
from salt.ext.tornado.ioloop import IOLoop
from salt.ext.tornado.netutil import ThreadedResolver

# When this module is imported, it runs getaddrinfo on a thread. Since
# the hostname is unicode, getaddrinfo attempts to import encodings.idna
# but blocks on the import lock. Verify that ThreadedResolver avoids
# this deadlock.

resolver = ThreadedResolver()
IOLoop.current().run_sync(lambda: resolver.resolve(u'localhost', 80))



import salt.states.kapacitor as kapacitor
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import Mock, mock_open, patch
from tests.support.unit import TestCase


def _present(
    name="testname",
    tick_script="/tmp/script.tick",
    task_type="stream",
    database="testdb",
    retention_policy="default",
    dbrps=None,
    enable=True,
    task=None,
    define_result=True,
    enable_result=True,
    disable_result=True,
    script="testscript",
):
    """
    Run a "kapacitor.present" state after setting up mocks, and return the
    state return value as well as the mocks to make assertions.
    """
    get_mock = Mock(return_value=task)

    if isinstance(define_result, bool):
        define_result = {"success": define_result}
    define_mock = Mock(return_value=define_result)

    if isinstance(enable_result, bool):
        enable_result = {"success": enable_result}
    enable_mock = Mock(return_value=enable_result)

    if isinstance(disable_result, bool):
        disable_result = {"success": disable_result}
    disable_mock = Mock(return_value=disable_result)

    with patch.dict(
        kapacitor.__salt__,
        {
            "kapacitor.get_task": get_mock,
            "kapacitor.define_task": define_mock,
            "kapacitor.enable_task": enable_mock,
            "kapacitor.disable_task": disable_mock,
        },
    ):
        with patch("salt.utils.files.fopen", mock_open(read_data=script)) as open_mock:
            retval = kapacitor.task_present(
                name,
                tick_script,
                task_type=task_type,
                database=database,
                retention_policy=retention_policy,
                enable=enable,
                dbrps=dbrps,
            )

    return retval, get_mock, define_mock, enable_mock, disable_mock


def _task(
    script="testscript", enabled=True, task_type="stream", db="testdb", rp="default"
):
    return {
        "script": script,
        "enabled": enabled,
        "type": task_type,
        "dbrps": [{"db": db, "rp": rp}],
    }


class KapacitorTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {kapacitor: {"__opts__": {"test": False}, "__env__": "test"}}

    def test_task_present_new_task(self):
        ret, get_mock, define_mock, enable_mock, _ = _present(
            dbrps=["testdb2.default_rp"]
        )
        get_mock.assert_called_once_with("testname")
        define_mock.assert_called_once_with(
            "testname",
            "/tmp/script.tick",
            database="testdb",
            retention_policy="default",
            task_type="stream",
            dbrps=["testdb2.default_rp", "testdb.default"],
        )
        enable_mock.assert_called_once_with("testname")
        self.assertIn("TICKscript diff", ret["changes"])
        self.assertIn("enabled", ret["changes"])
        self.assertEqual(True, ret["changes"]["enabled"]["new"])

    def test_task_present_existing_task_updated_script(self):
        ret, get_mock, define_mock, enable_mock, _ = _present(
            task=_task(script="oldscript")
        )
        get_mock.assert_called_once_with("testname")
        define_mock.assert_called_once_with(
            "testname",
            "/tmp/script.tick",
            database="testdb",
            retention_policy="default",
            task_type="stream",
            dbrps=["testdb.default"],
        )
        self.assertEqual(False, enable_mock.called)
        self.assertIn("TICKscript diff", ret["changes"])
        self.assertNotIn("enabled", ret["changes"])

    def test_task_present_existing_task_not_enabled(self):
        ret, get_mock, define_mock, enable_mock, _ = _present(task=_task(enabled=False))
        get_mock.assert_called_once_with("testname")
        self.assertEqual(False, define_mock.called)
        enable_mock.assert_called_once_with("testname")
        self.assertNotIn("diff", ret["changes"])
        self.assertIn("enabled", ret["changes"])
        self.assertEqual(True, ret["changes"]["enabled"]["new"])

    def test_task_present_disable_existing_task(self):
        ret, get_mock, define_mock, _, disable_mock = _present(
            task=_task(), enable=False
        )
        get_mock.assert_called_once_with("testname")
        self.assertEqual(False, define_mock.called)
        disable_mock.assert_called_once_with("testname")
        self.assertNotIn("diff", ret["changes"])
        self.assertIn("enabled", ret["changes"])
        self.assertEqual(False, ret["changes"]["enabled"]["new"])



"""
Take data from salt and "return" it into a raw file containing the json, with
one line per event.

Add the following to the minion or master configuration file.

.. code-block:: yaml

    rawfile_json.filename: <path_to_output_file>

Default is ``/var/log/salt/events``.

Common use is to log all events on the master. This can generate a lot of
noise, so you may wish to configure batch processing and/or configure the
:conf_master:`event_return_whitelist` or :conf_master:`event_return_blacklist`
to restrict the events that are written.
"""


import logging

import salt.returners
import salt.utils.files
import salt.utils.json

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "rawfile_json"


def __virtual__():
    return __virtualname__


def _get_options(ret):
    """
    Returns options used for the rawfile_json returner.
    """
    defaults = {"filename": "/var/log/salt/events"}
    attrs = {"filename": "filename"}
    _options = salt.returners.get_returner_options(
        __virtualname__,
        ret,
        attrs,
        __salt__=__salt__,
        __opts__=__opts__,
        defaults=defaults,
    )

    return _options


def returner(ret):
    """
    Write the return data to a file on the minion.
    """
    opts = _get_options(ret)
    try:
        with salt.utils.files.flopen(opts["filename"], "a") as logfile:
            salt.utils.json.dump(ret, logfile)
            logfile.write("\n")
    except Exception:  # pylint: disable=broad-except
        log.error("Could not write to rawdata_json file %s", opts["filename"])
        raise


def event_return(events):
    """
    Write event data (return data and non-return data) to file on the master.
    """
    if len(events) == 0:
        # events is an empty list.
        # Don't open the logfile in vain.
        return
    opts = _get_options({})  # Pass in empty ret, since this is a list of events
    try:
        with salt.utils.files.flopen(opts["filename"], "a") as logfile:
            for event in events:
                salt.utils.json.dump(event, logfile)
                logfile.write("\n")
    except Exception:  # pylint: disable=broad-except
        log.error("Could not write to rawdata_json file %s", opts["filename"])
        raise



"""
integration tests for mac_softwareupdate
"""

import pytest
from tests.support.case import ModuleCase
from tests.support.helpers import runs_on


@pytest.mark.skip_if_not_root
@runs_on(kernel="Darwin")
@pytest.mark.skip_if_binaries_missing("softwareupdate")
class MacSoftwareUpdateModuleTest(ModuleCase):
    """
    Validate the mac_softwareupdate module
    """

    IGNORED_LIST = []
    SCHEDULE = False
    CATALOG = ""

    def setUp(self):
        """
        Get current settings
        """
        self.IGNORED_LIST = self.run_function("softwareupdate.list_ignored")
        self.SCHEDULE = self.run_function("softwareupdate.schedule")
        self.CATALOG = self.run_function("softwareupdate.get_catalog")

        super().setUp()

    def tearDown(self):
        """
        Reset to original settings
        """
        if self.IGNORED_LIST:
            for item in self.IGNORED_LIST:
                self.run_function("softwareupdate.ignore", [item])
        else:
            self.run_function("softwareupdate.reset_ignored")

        self.run_function("softwareupdate.schedule", [self.SCHEDULE])

        if self.CATALOG == "Default":
            self.run_function("softwareupdate.reset_catalog")
        else:
            self.run_function("softwareupdate.set_catalog", [self.CATALOG])

        super().tearDown()

    @pytest.mark.slow_test
    def test_list_available(self):
        """
        Test softwareupdate.list_available
        """
        # Can't predict what will be returned, so can only test that the return
        # is the correct type, dict
        self.assertIsInstance(self.run_function("softwareupdate.list_available"), dict)

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_ignore(self):
        """
        Test softwareupdate.ignore
        Test softwareupdate.list_ignored
        Test softwareupdate.reset_ignored
        """
        # Test reset_ignored
        self.assertTrue(self.run_function("softwareupdate.reset_ignored"))
        self.assertEqual(self.run_function("softwareupdate.list_ignored"), [])

        # Test ignore
        self.assertTrue(self.run_function("softwareupdate.ignore", ["spongebob"]))
        self.assertTrue(self.run_function("softwareupdate.ignore", ["squidward"]))

        # Test list_ignored and verify ignore
        self.assertIn("spongebob", self.run_function("softwareupdate.list_ignored"))
        self.assertIn("squidward", self.run_function("softwareupdate.list_ignored"))

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_schedule(self):
        """
        Test softwareupdate.schedule_enable
        Test softwareupdate.schedule_enabled
        """
        # Test enable
        self.assertTrue(self.run_function("softwareupdate.schedule_enable", [True]))
        self.assertTrue(self.run_function("softwareupdate.schedule_enabled"))

        # Test disable in case it was already enabled
        self.assertTrue(self.run_function("softwareupdate.schedule_enable", [False]))
        self.assertFalse(self.run_function("softwareupdate.schedule_enabled"))

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_update(self):
        """
        Test softwareupdate.update_all
        Test softwareupdate.update
        Test softwareupdate.update_available

        Need to know the names of updates that are available to properly test
        the update functions...
        """
        # There's no way to know what the dictionary will contain, so all we can
        # check is that the return is a dictionary
        self.assertIsInstance(self.run_function("softwareupdate.update_all"), dict)

        # Test update_available
        self.assertFalse(
            self.run_function("softwareupdate.update_available", ["spongebob"])
        )

        # Test update not available
        self.assertIn(
            "Update not available",
            self.run_function("softwareupdate.update", ["spongebob"]),
        )

    @pytest.mark.slow_test
    def test_list_downloads(self):
        """
        Test softwareupdate.list_downloads
        """
        self.assertIsInstance(self.run_function("softwareupdate.list_downloads"), list)

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_download(self):
        """
        Test softwareupdate.download

        Need to know the names of updates that are available to properly test
        the download function
        """
        # Test update not available
        self.assertIn(
            "Update not available",
            self.run_function("softwareupdate.download", ["spongebob"]),
        )

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_download_all(self):
        """
        Test softwareupdate.download_all
        """
        self.assertIsInstance(self.run_function("softwareupdate.download_all"), list)

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_get_set_reset_catalog(self):
        """
        Test softwareupdate.download_all
        """
        # Reset the catalog
        self.assertTrue(self.run_function("softwareupdate.reset_catalog"))
        self.assertEqual(self.run_function("softwareupdate.get_catalog"), "Default")

        # Test setting and getting the catalog
        self.assertTrue(self.run_function("softwareupdate.set_catalog", ["spongebob"]))
        self.assertEqual(self.run_function("softwareupdate.get_catalog"), "spongebob")

        # Test reset the catalog
        self.assertTrue(self.run_function("softwareupdate.reset_catalog"))
        self.assertEqual(self.run_function("softwareupdate.get_catalog"), "Default")



from urllib.parse import urlencode

import salt.utils.json
import salt.utils.yaml
from tests.support.cherrypy_testclasses import BaseToolsTest


class TestOutFormats(BaseToolsTest):
    def __get_cp_config__(self):
        return {
            "tools.hypermedia_out.on": True,
        }

    def test_default_accept(self):
        request, response = self.request("/")
        self.assertEqual(response.headers["Content-type"], "application/json")

    def test_unsupported_accept(self):
        request, response = self.request(
            "/", headers=(("Accept", "application/ms-word"),)
        )
        self.assertEqual(response.status, "406 Not Acceptable")

    def test_json_out(self):
        request, response = self.request("/", headers=(("Accept", "application/json"),))
        self.assertEqual(response.headers["Content-type"], "application/json")

    def test_yaml_out(self):
        request, response = self.request(
            "/", headers=(("Accept", "application/x-yaml"),)
        )
        self.assertEqual(response.headers["Content-type"], "application/x-yaml")


class TestInFormats(BaseToolsTest):
    def __get_cp_config__(self):
        return {
            "tools.hypermedia_in.on": True,
        }

    def test_urlencoded_ctype(self):
        data = {"valid": "stuff"}
        raw = "valid=stuff"
        request, response = self.request(
            "/",
            method="POST",
            body=urlencode(data),
            headers=(("Content-type", "application/x-www-form-urlencoded"),),
        )
        self.assertEqual(response.status, "200 OK")
        self.assertEqual(request.raw_body, raw)
        self.assertDictEqual(request.unserialized_data, data)

    def test_urlencoded_multi_args(self):
        multi_args = "arg=arg1&arg=arg2"
        expected = {"arg": ["arg1", "arg2"]}
        request, response = self.request(
            "/",
            method="POST",
            body=multi_args,
            headers=(("Content-type", "application/x-www-form-urlencoded"),),
        )
        self.assertEqual(response.status, "200 OK")
        self.assertEqual(request.raw_body, multi_args)
        self.assertDictEqual(request.unserialized_data, expected)

    def test_json_ctype(self):
        data = {"valid": "stuff"}
        request, response = self.request(
            "/",
            method="POST",
            body=salt.utils.json.dumps(data),
            headers=(("Content-type", "application/json"),),
        )
        self.assertEqual(response.status, "200 OK")
        self.assertDictEqual(request.unserialized_data, data)

    def test_json_as_text_out(self):
        """
        Some service send JSON as text/plain for compatibility purposes
        """
        data = {"valid": "stuff"}
        request, response = self.request(
            "/",
            method="POST",
            body=salt.utils.json.dumps(data),
            headers=(("Content-type", "text/plain"),),
        )
        self.assertEqual(response.status, "200 OK")
        self.assertDictEqual(request.unserialized_data, data)

    def test_yaml_ctype(self):
        data = {"valid": "stuff"}
        request, response = self.request(
            "/",
            method="POST",
            body=salt.utils.yaml.safe_dump(data),
            headers=(("Content-type", "application/x-yaml"),),
        )
        self.assertEqual(response.status, "200 OK")
        self.assertDictEqual(request.unserialized_data, data)


class TestCors(BaseToolsTest):
    def __get_cp_config__(self):
        return {
            "tools.cors_tool.on": True,
        }

    def test_option_request(self):
        request, response = self.request(
            "/", method="OPTIONS", headers=(("Origin", "https://domain.com"),)
        )
        self.assertEqual(response.status, "200 OK")
        self.assertEqual(
            response.headers.get("Access-Control-Allow-Origin"), "https://domain.com"
        )



"""
Module for gathering and managing bridging information
"""

import re
import sys

import salt.utils.path

__func_alias__ = {"list_": "list"}


# Other BSD-like derivatives that use ifconfig may work too
SUPPORTED_BSD_LIKE = ["FreeBSD", "NetBSD", "OpenBSD"]


def __virtual__():
    """
    Confirm this module is supported by the OS and the system has
    required tools
    """
    supported_os_tool = {
        "FreeBSD": "ifconfig",
        "Linux": "brctl",
        "NetBSD": "brconfig",
        "OpenBSD": "ifconfig",
    }
    cur_os = __grains__["kernel"]
    for _os in supported_os_tool:
        if cur_os == _os and salt.utils.path.which(supported_os_tool[cur_os]):
            return True
    return (
        False,
        "The bridge execution module failed to load: requires one of the following"
        " tool/os combinations: ifconfig on FreeBSD/OpenBSD, brctl on Linux or brconfig"
        " on NetBSD.",
    )


def _tool_path(ostool):
    """
    Internal, returns tools path
    """
    return salt.utils.path.which(ostool)


def _linux_brshow(br=None):
    """
    Internal, returns bridges and enslaved interfaces (GNU/Linux - brctl)
    """
    brctl = _tool_path("brctl")

    if br:
        cmd = "{} show {}".format(brctl, br)
    else:
        cmd = "{} show".format(brctl)

    brs = {}

    for line in __salt__["cmd.run"](cmd, python_shell=False).splitlines():
        # get rid of first line
        if line.startswith("bridge name"):
            continue
        # get rid of ^\n's
        vals = line.split()
        if not vals:
            continue

        # bridge name bridge id       STP enabled interfaces
        # br0       8000.e4115bac8ddc   no      eth0
        #                                       foo0
        # br1       8000.e4115bac8ddc   no      eth1
        if len(vals) > 1:
            brname = vals[0]

            brs[brname] = {
                "id": vals[1],
                "stp": vals[2],
            }
            if len(vals) > 3:
                brs[brname]["interfaces"] = [vals[3]]

        if len(vals) == 1 and brname:
            brs[brname]["interfaces"].append(vals[0])

    if br:
        try:
            return brs[br]
        except KeyError:
            return None
    return brs


def _linux_bradd(br):
    """
    Internal, creates the bridge
    """
    brctl = _tool_path("brctl")
    return __salt__["cmd.run"]("{} addbr {}".format(brctl, br), python_shell=False)


def _linux_brdel(br):
    """
    Internal, deletes the bridge
    """
    brctl = _tool_path("brctl")
    return __salt__["cmd.run"]("{} delbr {}".format(brctl, br), python_shell=False)


def _linux_addif(br, iface):
    """
    Internal, adds an interface to a bridge
    """
    brctl = _tool_path("brctl")
    return __salt__["cmd.run"](
        "{} addif {} {}".format(brctl, br, iface), python_shell=False
    )


def _linux_delif(br, iface):
    """
    Internal, removes an interface from a bridge
    """
    brctl = _tool_path("brctl")
    return __salt__["cmd.run"](
        "{} delif {} {}".format(brctl, br, iface), python_shell=False
    )


def _linux_stp(br, state):
    """
    Internal, sets STP state
    """
    brctl = _tool_path("brctl")
    return __salt__["cmd.run"](
        "{} stp {} {}".format(brctl, br, state), python_shell=False
    )


def _bsd_brshow(br=None):
    """
    Internal, returns bridges and member interfaces (BSD-like: ifconfig)
    """
    if __grains__["kernel"] == "NetBSD":
        return _netbsd_brshow(br)

    ifconfig = _tool_path("ifconfig")

    ifaces = {}

    if br:
        ifaces[br] = br
    else:
        cmd = "{} -g bridge".format(ifconfig)
        for line in __salt__["cmd.run"](cmd, python_shell=False).splitlines():
            ifaces[line] = line

    brs = {}

    for iface in ifaces:
        cmd = "{} {}".format(ifconfig, iface)
        for line in __salt__["cmd.run"](cmd, python_shell=False).splitlines():
            brs[iface] = {"interfaces": [], "stp": "no"}
            line = line.lstrip()
            if line.startswith("member:"):
                brs[iface]["interfaces"].append(line.split(" ")[1])
                if "STP" in line:
                    brs[iface]["stp"] = "yes"

    if br:
        return brs[br]
    return brs


def _netbsd_brshow(br=None):
    """
    Internal, returns bridges and enslaved interfaces (NetBSD - brconfig)
    """
    brconfig = _tool_path("brconfig")

    if br:
        cmd = "{} {}".format(brconfig, br)
    else:
        cmd = "{} -a".format(brconfig)

    brs = {}
    start_int = False

    for line in __salt__["cmd.run"](cmd, python_shell=False).splitlines():
        if line.startswith("bridge"):
            start_int = False
            brname = line.split(":")[0]  # on NetBSD, always ^bridge([0-9]+):
            brs[brname] = {"interfaces": [], "stp": "no"}
        if "Interfaces:" in line:
            start_int = True
            continue
        if start_int and brname:
            m = re.match(r"\s*([a-z0-9]+)\s.*<.*>", line)
            if m:
                brs[brname]["interfaces"].append(m.group(1))
                if "STP" in line:
                    brs[brname]["stp"] = "yes"

    if br:
        try:
            return brs[br]
        except KeyError:
            return None
    return brs


def _bsd_bradd(br):
    """
    Internal, creates the bridge
    """
    kernel = __grains__["kernel"]
    ifconfig = _tool_path("ifconfig")

    if not br:
        return False

    if (
        __salt__["cmd.retcode"](
            "{} {} create up".format(ifconfig, br), python_shell=False
        )
        != 0
    ):
        return False

    # NetBSD is two cmds
    if kernel == "NetBSD":
        brconfig = _tool_path("brconfig")
        if (
            __salt__["cmd.retcode"]("{} {} up".format(brconfig, br), python_shell=False)
            != 0
        ):
            return False

    return True


def _bsd_brdel(br):
    """
    Internal, deletes the bridge
    """
    ifconfig = _tool_path("ifconfig")
    if not br:
        return False
    return __salt__["cmd.run"]("{} {} destroy".format(ifconfig, br), python_shell=False)


def _bsd_addif(br, iface):
    """
    Internal, adds an interface to a bridge
    """
    kernel = __grains__["kernel"]
    if kernel == "NetBSD":
        cmd = _tool_path("brconfig")
        brcmd = "add"
    else:
        cmd = _tool_path("ifconfig")
        brcmd = "addem"

    if not br or not iface:
        return False

    return __salt__["cmd.run"](
        "{} {} {} {}".format(cmd, br, brcmd, iface), python_shell=False
    )


def _bsd_delif(br, iface):
    """
    Internal, removes an interface from a bridge
    """
    kernel = __grains__["kernel"]
    if kernel == "NetBSD":
        cmd = _tool_path("brconfig")
        brcmd = "delete"
    else:
        cmd = _tool_path("ifconfig")
        brcmd = "deletem"

    if not br or not iface:
        return False

    return __salt__["cmd.run"](
        "{} {} {} {}".format(cmd, br, brcmd, iface), python_shell=False
    )


def _bsd_stp(br, state, iface):
    """
    Internal, sets STP state. On BSD-like, it is required to specify the
    STP physical interface
    """
    kernel = __grains__["kernel"]
    if kernel == "NetBSD":
        cmd = _tool_path("brconfig")
    else:
        cmd = _tool_path("ifconfig")

    if not br or not iface:
        return False

    return __salt__["cmd.run"](
        "{} {} {} {}".format(cmd, br, state, iface), python_shell=False
    )


def _os_dispatch(func, *args, **kwargs):
    """
    Internal, dispatches functions by operating system
    """
    if __grains__["kernel"] in SUPPORTED_BSD_LIKE:
        kernel = "bsd"
    else:
        kernel = __grains__["kernel"].lower()

    _os_func = getattr(sys.modules[__name__], "_{}_{}".format(kernel, func))

    if callable(_os_func):
        return _os_func(*args, **kwargs)


# End of internal functions


def show(br=None):
    """
    Returns bridges interfaces along with enslaved physical interfaces. If
    no interface is given, all bridges are shown, else only the specified
    bridge values are returned.

    CLI Example:

    .. code-block:: bash

        salt '*' bridge.show
        salt '*' bridge.show br0
    """
    return _os_dispatch("brshow", br)


def list_():
    """
    Returns the machine's bridges list

    CLI Example:

    .. code-block:: bash

        salt '*' bridge.list
    """
    brs = _os_dispatch("brshow")
    if not brs:
        return None
    brlist = []
    for br in brs:
        brlist.append(br)

    return brlist


def interfaces(br=None):
    """
    Returns interfaces attached to a bridge

    CLI Example:

    .. code-block:: bash

        salt '*' bridge.interfaces br0
    """
    if not br:
        return None

    br_ret = _os_dispatch("brshow", br)
    if br_ret:
        return br_ret["interfaces"]


def find_interfaces(*args):
    """
    Returns the bridge to which the interfaces are bond to

    CLI Example:

    .. code-block:: bash

        salt '*' bridge.find_interfaces eth0 [eth1...]
    """
    brs = _os_dispatch("brshow")
    if not brs:
        return None

    iflist = {}

    for iface in args:
        for br in brs:
            try:  # a bridge may not contain interfaces
                if iface in brs[br]["interfaces"]:
                    iflist[iface] = br
            except Exception:  # pylint: disable=broad-except
                pass

    return iflist


def add(br=None):
    """
    Creates a bridge

    CLI Example:

    .. code-block:: bash

        salt '*' bridge.add br0
    """
    return _os_dispatch("bradd", br)


def delete(br=None):
    """
    Deletes a bridge

    CLI Example:

    .. code-block:: bash

        salt '*' bridge.delete br0
    """
    return _os_dispatch("brdel", br)


def addif(br=None, iface=None):
    """
    Adds an interface to a bridge

    CLI Example:

    .. code-block:: bash

        salt '*' bridge.addif br0 eth0
    """
    return _os_dispatch("addif", br, iface)


def delif(br=None, iface=None):
    """
    Removes an interface from a bridge

    CLI Example:

    .. code-block:: bash

        salt '*' bridge.delif br0 eth0
    """
    return _os_dispatch("delif", br, iface)


def stp(br=None, state="disable", iface=None):
    """
    Sets Spanning Tree Protocol state for a bridge

    CLI Example:

    .. code-block:: bash

        salt '*' bridge.stp br0 enable
        salt '*' bridge.stp br0 disable

    For BSD-like operating systems, it is required to add the interface on
    which to enable the STP.

    CLI Example:

    .. code-block:: bash

        salt '*' bridge.stp bridge0 enable fxp0
        salt '*' bridge.stp bridge0 disable fxp0
    """
    kernel = __grains__["kernel"]
    if kernel == "Linux":
        states = {"enable": "on", "disable": "off"}
        return _os_dispatch("stp", br, states[state])
    elif kernel in SUPPORTED_BSD_LIKE:
        states = {"enable": "stp", "disable": "-stp"}
        return _os_dispatch("stp", br, states[state], iface)
    else:
        return False


# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



"""
Tests for the cron state
"""

import logging
import subprocess

import pytest
import salt.utils.platform

log = logging.getLogger(__name__)


@pytest.fixture
def cron_account():
    with pytest.helpers.create_account() as system_account:
        try:
            yield system_account
        finally:
            command = ["crontab", "-u", system_account.username, "-r"]
            if salt.utils.platform.is_freebsd():
                command.append("-f")
            subprocess.run(command, check=False)


@pytest.mark.slow_test
@pytest.mark.skip_on_windows
@pytest.mark.skip_if_not_root
@pytest.mark.skip_if_binaries_missing("crontab")
def test_managed(cron_account, salt_cli, salt_minion, base_env_state_tree_root_dir):
    """
    file.managed
    """
    cron_contents = (
        "# Lines below here are managed by Salt, do not edit\n@hourly touch"
        " /tmp/test-file\n"
    )
    expected = (
        "--- \n+++ \n@@ -1 +1,2 @@\n-\n+# Lines below here are managed by Salt, do not"
        " edit\n+@hourly touch /tmp/test-file\n"
    )
    with pytest.helpers.temp_file(
        "issue-46881/cron", cron_contents, base_env_state_tree_root_dir
    ):
        ret = salt_cli.run(
            "state.single",
            "cron.file",
            name="salt://issue-46881/cron",
            user=cron_account.username,
            minion_tgt=salt_minion.id,
        )
    assert ret.exitcode == 0, ret
    state = ret.json["cron_|-salt://issue-46881/cron_|-salt://issue-46881/cron_|-file"]
    assert "changes" in state
    assert "diff" in state["changes"]
    assert state["changes"]["diff"] == expected



"""``tornado.gen`` is a generator-based interface to make it easier to
work in an asynchronous environment.  Code using the ``gen`` module
is technically asynchronous, but it is written as a single generator
instead of a collection of separate functions.

For example, the following asynchronous handler:

.. testcode::

    class AsyncHandler(RequestHandler):
        @asynchronous
        def get(self):
            http_client = AsyncHTTPClient()
            http_client.fetch("http://example.com",
                              callback=self.on_fetch)

        def on_fetch(self, response):
            do_something_with_response(response)
            self.render("template.html")

.. testoutput::
   :hide:

could be written with ``gen`` as:

.. testcode::

    class GenAsyncHandler(RequestHandler):
        @gen.coroutine
        def get(self):
            http_client = AsyncHTTPClient()
            response = yield http_client.fetch("http://example.com")
            do_something_with_response(response)
            self.render("template.html")

.. testoutput::
   :hide:

Most asynchronous functions in Tornado return a `.Future`;
yielding this object returns its `~.Future.result`.

You can also yield a list or dict of ``Futures``, which will be
started at the same time and run in parallel; a list or dict of results will
be returned when they are all finished:

.. testcode::

    @gen.coroutine
    def get(self):
        http_client = AsyncHTTPClient()
        response1, response2 = yield [http_client.fetch(url1),
                                      http_client.fetch(url2)]
        response_dict = yield dict(response3=http_client.fetch(url3),
                                   response4=http_client.fetch(url4))
        response3 = response_dict['response3']
        response4 = response_dict['response4']

.. testoutput::
   :hide:

If the `~functools.singledispatch` library is available (standard in
Python 3.4, available via the `singledispatch
<https://pypi.python.org/pypi/singledispatch>`_ package on older
versions), additional types of objects may be yielded. Tornado includes
support for ``asyncio.Future`` and Twisted's ``Deferred`` class when
``tornado.platform.asyncio`` and ``tornado.platform.twisted`` are imported.
See the `convert_yielded` function to extend this mechanism.

.. versionchanged:: 3.2
   Dict support added.

.. versionchanged:: 4.1
   Support added for yielding ``asyncio`` Futures and Twisted Deferreds
   via ``singledispatch``.

"""
# pylint: skip-file
from __future__ import absolute_import, division, print_function

import collections
import functools
import itertools
import os
import sys
import textwrap
import types
import weakref

import salt.ext.tornado as tornado
from salt.ext.tornado.concurrent import Future, TracebackFuture, is_future, chain_future
from salt.ext.tornado.ioloop import IOLoop
from salt.ext.tornado.log import app_log
from salt.ext.tornado import stack_context
from salt.ext.tornado.util import PY3, raise_exc_info

try:
    try:
        # py34+
        from functools import singledispatch  # type: ignore
    except ImportError:
        from singledispatch import singledispatch  # backport
except ImportError:
    # In most cases, singledispatch is required (to avoid
    # difficult-to-diagnose problems in which the functionality
    # available differs depending on which invisble packages are
    # installed). However, in Google App Engine third-party
    # dependencies are more trouble so we allow this module to be
    # imported without it.
    if 'APPENGINE_RUNTIME' not in os.environ:
        raise
    singledispatch = None

try:
    try:
        # py35+
        from collections.abc import Generator as GeneratorType  # type: ignore
    except ImportError:
        from salt.ext.backports_abc import Generator as GeneratorType  # type: ignore

    try:
        # py35+
        from inspect import isawaitable  # type: ignore
    except ImportError:
        from salt.ext.backports_abc import isawaitable
except ImportError:
    if 'APPENGINE_RUNTIME' not in os.environ:
        raise
    from types import GeneratorType

    def isawaitable(x):  # type: ignore
        return False

if PY3:
    import builtins
else:
    import __builtin__ as builtins


class KeyReuseError(Exception):
    pass


class UnknownKeyError(Exception):
    pass


class LeakedCallbackError(Exception):
    pass


class BadYieldError(Exception):
    pass


class ReturnValueIgnoredError(Exception):
    pass


class TimeoutError(Exception):
    """Exception raised by ``with_timeout``."""


def _value_from_stopiteration(e):
    try:
        # StopIteration has a value attribute beginning in py33.
        # So does our Return class.
        return e.value
    except AttributeError:
        pass
    try:
        # Cython backports coroutine functionality by putting the value in
        # e.args[0].
        return e.args[0]
    except (AttributeError, IndexError):
        return None


def engine(func):
    """Callback-oriented decorator for asynchronous generators.

    This is an older interface; for new code that does not need to be
    compatible with versions of Tornado older than 3.0 the
    `coroutine` decorator is recommended instead.

    This decorator is similar to `coroutine`, except it does not
    return a `.Future` and the ``callback`` argument is not treated
    specially.

    In most cases, functions decorated with `engine` should take
    a ``callback`` argument and invoke it with their result when
    they are finished.  One notable exception is the
    `~tornado.web.RequestHandler` :ref:`HTTP verb methods <verbs>`,
    which use ``self.finish()`` in place of a callback argument.
    """
    func = _make_coroutine_wrapper(func, replace_callback=False)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        future = func(*args, **kwargs)

        def final_callback(future):
            if future.result() is not None:
                raise ReturnValueIgnoredError(
                    "@gen.engine functions cannot return values: %r" %
                    (future.result(),))
        # The engine interface doesn't give us any way to return
        # errors but to raise them into the stack context.
        # Save the stack context here to use when the Future has resolved.
        future.add_done_callback(stack_context.wrap(final_callback))
    return wrapper


def coroutine(func, replace_callback=True):
    """Decorator for asynchronous generators.

    Any generator that yields objects from this module must be wrapped
    in either this decorator or `engine`.

    Coroutines may "return" by raising the special exception
    `Return(value) <Return>`.  In Python 3.3+, it is also possible for
    the function to simply use the ``return value`` statement (prior to
    Python 3.3 generators were not allowed to also return values).
    In all versions of Python a coroutine that simply wishes to exit
    early may use the ``return`` statement without a value.

    Functions with this decorator return a `.Future`.  Additionally,
    they may be called with a ``callback`` keyword argument, which
    will be invoked with the future's result when it resolves.  If the
    coroutine fails, the callback will not be run and an exception
    will be raised into the surrounding `.StackContext`.  The
    ``callback`` argument is not visible inside the decorated
    function; it is handled by the decorator itself.

    From the caller's perspective, ``@gen.coroutine`` is similar to
    the combination of ``@return_future`` and ``@gen.engine``.

    .. warning::

       When exceptions occur inside a coroutine, the exception
       information will be stored in the `.Future` object. You must
       examine the result of the `.Future` object, or the exception
       may go unnoticed by your code. This means yielding the function
       if called from another coroutine, using something like
       `.IOLoop.run_sync` for top-level calls, or passing the `.Future`
       to `.IOLoop.add_future`.

    """
    return _make_coroutine_wrapper(func, replace_callback=True)


# Ties lifetime of runners to their result futures. Github Issue #1769
# Generators, like any object in Python, must be strong referenced
# in order to not be cleaned up by the garbage collector. When using
# coroutines, the Runner object is what strong-refs the inner
# generator. However, the only item that strong-reffed the Runner
# was the last Future that the inner generator yielded (via the
# Future's internal done_callback list). Usually this is enough, but
# it is also possible for this Future to not have any strong references
# other than other objects referenced by the Runner object (usually
# when using other callback patterns and/or weakrefs). In this
# situation, if a garbage collection ran, a cycle would be detected and
# Runner objects could be destroyed along with their inner generators
# and everything in their local scope.
# This map provides strong references to Runner objects as long as
# their result future objects also have strong references (typically
# from the parent coroutine's Runner). This keeps the coroutine's
# Runner alive.
_futures_to_runners = weakref.WeakKeyDictionary()


def _make_coroutine_wrapper(func, replace_callback):
    """The inner workings of ``@gen.coroutine`` and ``@gen.engine``.

    The two decorators differ in their treatment of the ``callback``
    argument, so we cannot simply implement ``@engine`` in terms of
    ``@coroutine``.
    """
    # On Python 3.5, set the coroutine flag on our generator, to allow it
    # to be used with 'await'.
    wrapped = func
    if hasattr(types, 'coroutine'):
        func = types.coroutine(func)

    @functools.wraps(wrapped)
    def wrapper(*args, **kwargs):
        future = TracebackFuture()

        if replace_callback and 'callback' in kwargs:
            callback = kwargs.pop('callback')
            IOLoop.current().add_future(
                future, lambda future: callback(future.result()))

        try:
            result = func(*args, **kwargs)
        except (Return, StopIteration) as e:
            result = _value_from_stopiteration(e)
        except Exception:
            future.set_exc_info(sys.exc_info())
            return future
        else:
            if isinstance(result, GeneratorType):
                # Inline the first iteration of Runner.run.  This lets us
                # avoid the cost of creating a Runner when the coroutine
                # never actually yields, which in turn allows us to
                # use "optional" coroutines in critical path code without
                # performance penalty for the synchronous case.
                try:
                    orig_stack_contexts = stack_context._state.contexts
                    yielded = next(result)
                    if stack_context._state.contexts is not orig_stack_contexts:
                        yielded = TracebackFuture()
                        yielded.set_exception(
                            stack_context.StackContextInconsistentError(
                                'stack_context inconsistency (probably caused '
                                'by yield within a "with StackContext" block)'))
                except (StopIteration, Return) as e:
                    future.set_result(_value_from_stopiteration(e))
                except Exception:
                    future.set_exc_info(sys.exc_info())
                else:
                    _futures_to_runners[future] = Runner(result, future, yielded)
                yielded = None
                try:
                    return future
                finally:
                    # Subtle memory optimization: if next() raised an exception,
                    # the future's exc_info contains a traceback which
                    # includes this stack frame.  This creates a cycle,
                    # which will be collected at the next full GC but has
                    # been shown to greatly increase memory usage of
                    # benchmarks (relative to the refcount-based scheme
                    # used in the absence of cycles).  We can avoid the
                    # cycle by clearing the local variable after we return it.
                    future = None
        future.set_result(result)
        return future

    wrapper.__wrapped__ = wrapped
    wrapper.__tornado_coroutine__ = True
    return wrapper


def is_coroutine_function(func):
    """Return whether *func* is a coroutine function, i.e. a function
    wrapped with `~.gen.coroutine`.

    .. versionadded:: 4.5
    """
    return getattr(func, '__tornado_coroutine__', False)


class Return(Exception):
    """Special exception to return a value from a `coroutine`.

    If this exception is raised, its value argument is used as the
    result of the coroutine::

        @gen.coroutine
        def fetch_json(url):
            response = yield AsyncHTTPClient().fetch(url)
            raise gen.Return(json_decode(response.body))

    In Python 3.3, this exception is no longer necessary: the ``return``
    statement can be used directly to return a value (previously
    ``yield`` and ``return`` with a value could not be combined in the
    same function).

    By analogy with the return statement, the value argument is optional,
    but it is never necessary to ``raise gen.Return()``.  The ``return``
    statement can be used with no arguments instead.
    """
    def __init__(self, value=None):
        super(Return, self).__init__()
        self.value = value
        # Cython recognizes subclasses of StopIteration with a .args tuple.
        self.args = (value,)


class WaitIterator(object):
    """Provides an iterator to yield the results of futures as they finish.

    Yielding a set of futures like this:

    ``results = yield [future1, future2]``

    pauses the coroutine until both ``future1`` and ``future2``
    return, and then restarts the coroutine with the results of both
    futures. If either future is an exception, the expression will
    raise that exception and all the results will be lost.

    If you need to get the result of each future as soon as possible,
    or if you need the result of some futures even if others produce
    errors, you can use ``WaitIterator``::

      wait_iterator = gen.WaitIterator(future1, future2)
      while not wait_iterator.done():
          try:
              result = yield wait_iterator.next()
          except Exception as e:
              print("Error {} from {}".format(e, wait_iterator.current_future))
          else:
              print("Result {} received from {} at {}".format(
                  result, wait_iterator.current_future,
                  wait_iterator.current_index))

    Because results are returned as soon as they are available the
    output from the iterator *will not be in the same order as the
    input arguments*. If you need to know which future produced the
    current result, you can use the attributes
    ``WaitIterator.current_future``, or ``WaitIterator.current_index``
    to get the index of the future from the input list. (if keyword
    arguments were used in the construction of the `WaitIterator`,
    ``current_index`` will use the corresponding keyword).

    On Python 3.5, `WaitIterator` implements the async iterator
    protocol, so it can be used with the ``async for`` statement (note
    that in this version the entire iteration is aborted if any value
    raises an exception, while the previous example can continue past
    individual errors)::

      async for result in gen.WaitIterator(future1, future2):
          print("Result {} received from {} at {}".format(
              result, wait_iterator.current_future,
              wait_iterator.current_index))

    .. versionadded:: 4.1

    .. versionchanged:: 4.3
       Added ``async for`` support in Python 3.5.

    """
    def __init__(self, *args, **kwargs):
        if args and kwargs:
            raise ValueError(
                "You must provide args or kwargs, not both")

        if kwargs:
            self._unfinished = dict((f, k) for (k, f) in kwargs.items())
            futures = list(kwargs.values())
        else:
            self._unfinished = dict((f, i) for (i, f) in enumerate(args))
            futures = args

        self._finished = collections.deque()
        self.current_index = self.current_future = None
        self._running_future = None

        for future in futures:
            future.add_done_callback(self._done_callback)

    def done(self):
        """Returns True if this iterator has no more results."""
        if self._finished or self._unfinished:
            return False
        # Clear the 'current' values when iteration is done.
        self.current_index = self.current_future = None
        return True

    def next(self):
        """Returns a `.Future` that will yield the next available result.

        Note that this `.Future` will not be the same object as any of
        the inputs.
        """
        self._running_future = TracebackFuture()

        if self._finished:
            self._return_result(self._finished.popleft())

        return self._running_future

    def _done_callback(self, done):
        if self._running_future and not self._running_future.done():
            self._return_result(done)
        else:
            self._finished.append(done)

    def _return_result(self, done):
        """Called set the returned future's state that of the future
        we yielded, and set the current future for the iterator.
        """
        chain_future(done, self._running_future)

        self.current_future = done
        self.current_index = self._unfinished.pop(done)

    def __aiter__(self):
        raise self

    def __anext__(self):
        if self.done():
            # Lookup by name to silence pyflakes on older versions.
            raise getattr(builtins, 'StopAsyncIteration')()
        return self.next()


class YieldPoint(object):
    """Base class for objects that may be yielded from the generator.

    .. deprecated:: 4.0
       Use `Futures <.Future>` instead.
    """
    def start(self, runner):
        """Called by the runner after the generator has yielded.

        No other methods will be called on this object before ``start``.
        """
        raise NotImplementedError()

    def is_ready(self):
        """Called by the runner to determine whether to resume the generator.

        Returns a boolean; may be called more than once.
        """
        raise NotImplementedError()

    def get_result(self):
        """Returns the value to use as the result of the yield expression.

        This method will only be called once, and only after `is_ready`
        has returned true.
        """
        raise NotImplementedError()


class Callback(YieldPoint):
    """Returns a callable object that will allow a matching `Wait` to proceed.

    The key may be any value suitable for use as a dictionary key, and is
    used to match ``Callbacks`` to their corresponding ``Waits``.  The key
    must be unique among outstanding callbacks within a single run of the
    generator function, but may be reused across different runs of the same
    function (so constants generally work fine).

    The callback may be called with zero or one arguments; if an argument
    is given it will be returned by `Wait`.

    .. deprecated:: 4.0
       Use `Futures <.Future>` instead.
    """
    def __init__(self, key):
        self.key = key

    def start(self, runner):
        self.runner = runner
        runner.register_callback(self.key)

    def is_ready(self):
        return True

    def get_result(self):
        return self.runner.result_callback(self.key)


class Wait(YieldPoint):
    """Returns the argument passed to the result of a previous `Callback`.

    .. deprecated:: 4.0
       Use `Futures <.Future>` instead.
    """
    def __init__(self, key):
        self.key = key

    def start(self, runner):
        self.runner = runner

    def is_ready(self):
        return self.runner.is_ready(self.key)

    def get_result(self):
        return self.runner.pop_result(self.key)


class WaitAll(YieldPoint):
    """Returns the results of multiple previous `Callbacks <Callback>`.

    The argument is a sequence of `Callback` keys, and the result is
    a list of results in the same order.

    `WaitAll` is equivalent to yielding a list of `Wait` objects.

    .. deprecated:: 4.0
       Use `Futures <.Future>` instead.
    """
    def __init__(self, keys):
        self.keys = keys

    def start(self, runner):
        self.runner = runner

    def is_ready(self):
        return all(self.runner.is_ready(key) for key in self.keys)

    def get_result(self):
        return [self.runner.pop_result(key) for key in self.keys]


def Task(func, *args, **kwargs):
    """Adapts a callback-based asynchronous function for use in coroutines.

    Takes a function (and optional additional arguments) and runs it with
    those arguments plus a ``callback`` keyword argument.  The argument passed
    to the callback is returned as the result of the yield expression.

    .. versionchanged:: 4.0
       ``gen.Task`` is now a function that returns a `.Future`, instead of
       a subclass of `YieldPoint`.  It still behaves the same way when
       yielded.
    """
    future = Future()

    def handle_exception(typ, value, tb):
        if future.done():
            return False
        future.set_exc_info((typ, value, tb))
        return True

    def set_result(result):
        if future.done():
            return
        future.set_result(result)
    with stack_context.ExceptionStackContext(handle_exception):
        func(*args, callback=_argument_adapter(set_result), **kwargs)
    return future


class YieldFuture(YieldPoint):
    def __init__(self, future, io_loop=None):
        """Adapts a `.Future` to the `YieldPoint` interface.

        .. versionchanged:: 4.1
           The ``io_loop`` argument is deprecated.
        """
        self.future = future
        self.io_loop = io_loop or IOLoop.current()

    def start(self, runner):
        if not self.future.done():
            self.runner = runner
            self.key = object()
            runner.register_callback(self.key)
            self.io_loop.add_future(self.future, runner.result_callback(self.key))
        else:
            self.runner = None
            self.result_fn = self.future.result

    def is_ready(self):
        if self.runner is not None:
            return self.runner.is_ready(self.key)
        else:
            return True

    def get_result(self):
        if self.runner is not None:
            return self.runner.pop_result(self.key).result()
        else:
            return self.result_fn()


def _contains_yieldpoint(children):
    """Returns True if ``children`` contains any YieldPoints.

    ``children`` may be a dict or a list, as used by `MultiYieldPoint`
    and `multi_future`.
    """
    if isinstance(children, dict):
        return any(isinstance(i, YieldPoint) for i in children.values())
    if isinstance(children, list):
        return any(isinstance(i, YieldPoint) for i in children)
    return False


def multi(children, quiet_exceptions=()):
    """Runs multiple asynchronous operations in parallel.

    ``children`` may either be a list or a dict whose values are
    yieldable objects. ``multi()`` returns a new yieldable
    object that resolves to a parallel structure containing their
    results. If ``children`` is a list, the result is a list of
    results in the same order; if it is a dict, the result is a dict
    with the same keys.

    That is, ``results = yield multi(list_of_futures)`` is equivalent
    to::

        results = []
        for future in list_of_futures:
            results.append(yield future)

    If any children raise exceptions, ``multi()`` will raise the first
    one. All others will be logged, unless they are of types
    contained in the ``quiet_exceptions`` argument.

    If any of the inputs are `YieldPoints <YieldPoint>`, the returned
    yieldable object is a `YieldPoint`. Otherwise, returns a `.Future`.
    This means that the result of `multi` can be used in a native
    coroutine if and only if all of its children can be.

    In a ``yield``-based coroutine, it is not normally necessary to
    call this function directly, since the coroutine runner will
    do it automatically when a list or dict is yielded. However,
    it is necessary in ``await``-based coroutines, or to pass
    the ``quiet_exceptions`` argument.

    This function is available under the names ``multi()`` and ``Multi()``
    for historical reasons.

    .. versionchanged:: 4.2
       If multiple yieldables fail, any exceptions after the first
       (which is raised) will be logged. Added the ``quiet_exceptions``
       argument to suppress this logging for selected exception types.

    .. versionchanged:: 4.3
       Replaced the class ``Multi`` and the function ``multi_future``
       with a unified function ``multi``. Added support for yieldables
       other than `YieldPoint` and `.Future`.

    """
    if _contains_yieldpoint(children):
        return MultiYieldPoint(children, quiet_exceptions=quiet_exceptions)
    else:
        return multi_future(children, quiet_exceptions=quiet_exceptions)


Multi = multi


class MultiYieldPoint(YieldPoint):
    """Runs multiple asynchronous operations in parallel.

    This class is similar to `multi`, but it always creates a stack
    context even when no children require it. It is not compatible with
    native coroutines.

    .. versionchanged:: 4.2
       If multiple ``YieldPoints`` fail, any exceptions after the first
       (which is raised) will be logged. Added the ``quiet_exceptions``
       argument to suppress this logging for selected exception types.

    .. versionchanged:: 4.3
       Renamed from ``Multi`` to ``MultiYieldPoint``. The name ``Multi``
       remains as an alias for the equivalent `multi` function.

    .. deprecated:: 4.3
       Use `multi` instead.
    """
    def __init__(self, children, quiet_exceptions=()):
        self.keys = None
        if isinstance(children, dict):
            self.keys = list(children.keys())
            children = children.values()
        self.children = []
        for i in children:
            if not isinstance(i, YieldPoint):
                i = convert_yielded(i)
            if is_future(i):
                i = YieldFuture(i)
            self.children.append(i)
        assert all(isinstance(i, YieldPoint) for i in self.children)
        self.unfinished_children = set(self.children)
        self.quiet_exceptions = quiet_exceptions

    def start(self, runner):
        for i in self.children:
            i.start(runner)

    def is_ready(self):
        finished = list(itertools.takewhile(
            lambda i: i.is_ready(), self.unfinished_children))
        self.unfinished_children.difference_update(finished)
        return not self.unfinished_children

    def get_result(self):
        result_list = []
        exc_info = None
        for f in self.children:
            try:
                result_list.append(f.get_result())
            except Exception as e:
                if exc_info is None:
                    exc_info = sys.exc_info()
                else:
                    if not isinstance(e, self.quiet_exceptions):
                        app_log.error("Multiple exceptions in yield list",
                                      exc_info=True)
        if exc_info is not None:
            raise_exc_info(exc_info)
        if self.keys is not None:
            return dict(zip(self.keys, result_list))
        else:
            return list(result_list)


def multi_future(children, quiet_exceptions=()):
    """Wait for multiple asynchronous futures in parallel.

    This function is similar to `multi`, but does not support
    `YieldPoints <YieldPoint>`.

    .. versionadded:: 4.0

    .. versionchanged:: 4.2
       If multiple ``Futures`` fail, any exceptions after the first (which is
       raised) will be logged. Added the ``quiet_exceptions``
       argument to suppress this logging for selected exception types.

    .. deprecated:: 4.3
       Use `multi` instead.
    """
    if isinstance(children, dict):
        keys = list(children.keys())
        children = children.values()
    else:
        keys = None
    children = list(map(convert_yielded, children))
    assert all(is_future(i) for i in children)
    unfinished_children = set(children)

    future = Future()
    if not children:
        future.set_result({} if keys is not None else [])

    def callback(f):
        unfinished_children.remove(f)
        if not unfinished_children:
            result_list = []
            for f in children:
                try:
                    result_list.append(f.result())
                except Exception as e:
                    if future.done():
                        if not isinstance(e, quiet_exceptions):
                            app_log.error("Multiple exceptions in yield list",
                                          exc_info=True)
                    else:
                        future.set_exc_info(sys.exc_info())
            if not future.done():
                if keys is not None:
                    future.set_result(dict(zip(keys, result_list)))
                else:
                    future.set_result(result_list)

    listening = set()
    for f in children:
        if f not in listening:
            listening.add(f)
            f.add_done_callback(callback)
    return future


def maybe_future(x):
    """Converts ``x`` into a `.Future`.

    If ``x`` is already a `.Future`, it is simply returned; otherwise
    it is wrapped in a new `.Future`.  This is suitable for use as
    ``result = yield gen.maybe_future(f())`` when you don't know whether
    ``f()`` returns a `.Future` or not.

    .. deprecated:: 4.3
       This function only handles ``Futures``, not other yieldable objects.
       Instead of `maybe_future`, check for the non-future result types
       you expect (often just ``None``), and ``yield`` anything unknown.
    """
    if is_future(x):
        return x
    else:
        fut = Future()
        fut.set_result(x)
        return fut


def with_timeout(timeout, future, io_loop=None, quiet_exceptions=()):
    """Wraps a `.Future` (or other yieldable object) in a timeout.

    Raises `TimeoutError` if the input future does not complete before
    ``timeout``, which may be specified in any form allowed by
    `.IOLoop.add_timeout` (i.e. a `datetime.timedelta` or an absolute time
    relative to `.IOLoop.time`)

    If the wrapped `.Future` fails after it has timed out, the exception
    will be logged unless it is of a type contained in ``quiet_exceptions``
    (which may be an exception type or a sequence of types).

    Does not support `YieldPoint` subclasses.

    .. versionadded:: 4.0

    .. versionchanged:: 4.1
       Added the ``quiet_exceptions`` argument and the logging of unhandled
       exceptions.

    .. versionchanged:: 4.4
       Added support for yieldable objects other than `.Future`.
    """
    # TODO: allow YieldPoints in addition to other yieldables?
    # Tricky to do with stack_context semantics.
    #
    # It's tempting to optimize this by cancelling the input future on timeout
    # instead of creating a new one, but A) we can't know if we are the only
    # one waiting on the input future, so cancelling it might disrupt other
    # callers and B) concurrent futures can only be cancelled while they are
    # in the queue, so cancellation cannot reliably bound our waiting time.
    future = convert_yielded(future)
    result = Future()
    chain_future(future, result)
    if io_loop is None:
        io_loop = IOLoop.current()

    def error_callback(future):
        try:
            future.result()
        except Exception as e:
            if not isinstance(e, quiet_exceptions):
                app_log.error("Exception in Future %r after timeout",
                              future, exc_info=True)

    def timeout_callback():
        result.set_exception(TimeoutError("Timeout"))
        # In case the wrapped future goes on to fail, log it.
        future.add_done_callback(error_callback)
    timeout_handle = io_loop.add_timeout(
        timeout, timeout_callback)
    if isinstance(future, Future):
        # We know this future will resolve on the IOLoop, so we don't
        # need the extra thread-safety of IOLoop.add_future (and we also
        # don't care about StackContext here.
        future.add_done_callback(
            lambda future: io_loop.remove_timeout(timeout_handle))
    else:
        # concurrent.futures.Futures may resolve on any thread, so we
        # need to route them back to the IOLoop.
        io_loop.add_future(
            future, lambda future: io_loop.remove_timeout(timeout_handle))
    return result


def sleep(duration):
    """Return a `.Future` that resolves after the given number of seconds.

    When used with ``yield`` in a coroutine, this is a non-blocking
    analogue to `time.sleep` (which should not be used in coroutines
    because it is blocking)::

        yield gen.sleep(0.5)

    Note that calling this function on its own does nothing; you must
    wait on the `.Future` it returns (usually by yielding it).

    .. versionadded:: 4.1
    """
    f = Future()
    IOLoop.current().call_later(duration, lambda: f.set_result(None))
    return f


_null_future = Future()
_null_future.set_result(None)

moment = Future()
moment.__doc__ = \
    """A special object which may be yielded to allow the IOLoop to run for
one iteration.

This is not needed in normal use but it can be helpful in long-running
coroutines that are likely to yield Futures that are ready instantly.

Usage: ``yield gen.moment``

.. versionadded:: 4.0

.. deprecated:: 4.5
   ``yield None`` is now equivalent to ``yield gen.moment``.
"""
moment.set_result(None)


class Runner(object):
    """Internal implementation of `tornado.gen.engine`.

    Maintains information about pending callbacks and their results.

    The results of the generator are stored in ``result_future`` (a
    `.TracebackFuture`)
    """
    def __init__(self, gen, result_future, first_yielded):
        self.gen = gen
        self.result_future = result_future
        self.future = _null_future
        self.yield_point = None
        self.pending_callbacks = None
        self.results = None
        self.running = False
        self.finished = False
        self.had_exception = False
        self.io_loop = IOLoop.current()
        # For efficiency, we do not create a stack context until we
        # reach a YieldPoint (stack contexts are required for the historical
        # semantics of YieldPoints, but not for Futures).  When we have
        # done so, this field will be set and must be called at the end
        # of the coroutine.
        self.stack_context_deactivate = None
        if self.handle_yield(first_yielded):
            gen = result_future = first_yielded = None
            self.run()

    def register_callback(self, key):
        """Adds ``key`` to the list of callbacks."""
        if self.pending_callbacks is None:
            # Lazily initialize the old-style YieldPoint data structures.
            self.pending_callbacks = set()
            self.results = {}
        if key in self.pending_callbacks:
            raise KeyReuseError("key %r is already pending" % (key,))
        self.pending_callbacks.add(key)

    def is_ready(self, key):
        """Returns true if a result is available for ``key``."""
        if self.pending_callbacks is None or key not in self.pending_callbacks:
            raise UnknownKeyError("key %r is not pending" % (key,))
        return key in self.results

    def set_result(self, key, result):
        """Sets the result for ``key`` and attempts to resume the generator."""
        self.results[key] = result
        if self.yield_point is not None and self.yield_point.is_ready():
            try:
                self.future.set_result(self.yield_point.get_result())
            except:
                self.future.set_exc_info(sys.exc_info())
            self.yield_point = None
            self.run()

    def pop_result(self, key):
        """Returns the result for ``key`` and unregisters it."""
        self.pending_callbacks.remove(key)
        return self.results.pop(key)

    def run(self):
        """Starts or resumes the generator, running until it reaches a
        yield point that is not ready.
        """
        if self.running or self.finished:
            return
        try:
            self.running = True
            while True:
                future = self.future
                if not future.done():
                    return
                self.future = None
                try:
                    orig_stack_contexts = stack_context._state.contexts
                    exc_info = None

                    try:
                        value = future.result()
                    except Exception:
                        self.had_exception = True
                        exc_info = sys.exc_info()
                    future = None

                    if exc_info is not None:
                        try:
                            yielded = self.gen.throw(*exc_info)
                        finally:
                            # Break up a reference to itself
                            # for faster GC on CPython.
                            exc_info = None
                    else:
                        yielded = self.gen.send(value)

                    if stack_context._state.contexts is not orig_stack_contexts:
                        self.gen.throw(
                            stack_context.StackContextInconsistentError(
                                'stack_context inconsistency (probably caused '
                                'by yield within a "with StackContext" block)'))
                except (StopIteration, Return) as e:
                    self.finished = True
                    self.future = _null_future
                    if self.pending_callbacks and not self.had_exception:
                        # If we ran cleanly without waiting on all callbacks
                        # raise an error (really more of a warning).  If we
                        # had an exception then some callbacks may have been
                        # orphaned, so skip the check in that case.
                        raise LeakedCallbackError(
                            "finished without waiting for callbacks %r" %
                            self.pending_callbacks)
                    self.result_future.set_result(_value_from_stopiteration(e))
                    self.result_future = None
                    self._deactivate_stack_context()
                    return
                except Exception:
                    self.finished = True
                    self.future = _null_future
                    self.result_future.set_exc_info(sys.exc_info())
                    self.result_future = None
                    self._deactivate_stack_context()
                    return
                if not self.handle_yield(yielded):
                    return
                yielded = None
        finally:
            self.running = False

    def handle_yield(self, yielded):
        # Lists containing YieldPoints require stack contexts;
        # other lists are handled in convert_yielded.
        if _contains_yieldpoint(yielded):
            yielded = multi(yielded)

        if isinstance(yielded, YieldPoint):
            # YieldPoints are too closely coupled to the Runner to go
            # through the generic convert_yielded mechanism.
            self.future = TracebackFuture()

            def start_yield_point():
                try:
                    yielded.start(self)
                    if yielded.is_ready():
                        self.future.set_result(
                            yielded.get_result())
                    else:
                        self.yield_point = yielded
                except Exception:
                    self.future = TracebackFuture()
                    self.future.set_exc_info(sys.exc_info())

            if self.stack_context_deactivate is None:
                # Start a stack context if this is the first
                # YieldPoint we've seen.
                with stack_context.ExceptionStackContext(
                        self.handle_exception) as deactivate:
                    self.stack_context_deactivate = deactivate

                    def cb():
                        start_yield_point()
                        self.run()
                    self.io_loop.add_callback(cb)
                    return False
            else:
                start_yield_point()
        else:
            try:
                self.future = convert_yielded(yielded)
            except BadYieldError:
                self.future = TracebackFuture()
                self.future.set_exc_info(sys.exc_info())

        if not self.future.done() or self.future is moment:
            def inner(f):
                # Break a reference cycle to speed GC.
                f = None # noqa
                self.run()
            self.io_loop.add_future(
                self.future, inner)
            return False
        return True

    def result_callback(self, key):
        return stack_context.wrap(_argument_adapter(
            functools.partial(self.set_result, key)))

    def handle_exception(self, typ, value, tb):
        if not self.running and not self.finished:
            self.future = TracebackFuture()
            self.future.set_exc_info((typ, value, tb))
            self.run()
            return True
        else:
            return False

    def _deactivate_stack_context(self):
        if self.stack_context_deactivate is not None:
            self.stack_context_deactivate()
            self.stack_context_deactivate = None


Arguments = collections.namedtuple('Arguments', ['args', 'kwargs'])


def _argument_adapter(callback):
    """Returns a function that when invoked runs ``callback`` with one arg.

    If the function returned by this function is called with exactly
    one argument, that argument is passed to ``callback``.  Otherwise
    the args tuple and kwargs dict are wrapped in an `Arguments` object.
    """
    def wrapper(*args, **kwargs):
        if kwargs or len(args) > 1:
            callback(Arguments(args, kwargs))
        elif args:
            callback(args[0])
        else:
            callback(None)
    return wrapper


# Convert Awaitables into Futures. It is unfortunately possible
# to have infinite recursion here if those Awaitables assume that
# we're using a different coroutine runner and yield objects
# we don't understand. If that happens, the solution is to
# register that runner's yieldable objects with convert_yielded.
if sys.version_info >= (3, 3):
    exec(textwrap.dedent("""
    @coroutine
    def _wrap_awaitable(x):
        if hasattr(x, '__await__'):
            x = x.__await__()
        return (yield from x)
    """))
else:
    # Py2-compatible version for use with Cython.
    # Copied from PEP 380.
    @coroutine
    def _wrap_awaitable(x):
        if hasattr(x, '__await__'):
            _i = x.__await__()
        else:
            _i = iter(x)
        try:
            _y = next(_i)
        except StopIteration as _e:
            _r = _value_from_stopiteration(_e)
        else:
            while 1:
                try:
                    _s = yield _y
                except GeneratorExit as _e:
                    try:
                        _m = _i.close
                    except AttributeError:
                        pass
                    else:
                        _m()
                    raise _e
                except BaseException as _e:
                    _x = sys.exc_info()
                    try:
                        _m = _i.throw
                    except AttributeError:
                        raise _e
                    else:
                        try:
                            _y = _m(*_x)
                        except StopIteration as _e:
                            _r = _value_from_stopiteration(_e)
                            break
                else:
                    try:
                        if _s is None:
                            _y = next(_i)
                        else:
                            _y = _i.send(_s)
                    except StopIteration as _e:
                        _r = _value_from_stopiteration(_e)
                        break
        raise Return(_r)


def convert_yielded(yielded):
    """Convert a yielded object into a `.Future`.

    The default implementation accepts lists, dictionaries, and Futures.

    If the `~functools.singledispatch` library is available, this function
    may be extended to support additional types. For example::

        @convert_yielded.register(asyncio.Future)
        def _(asyncio_future):
            return tornado.platform.asyncio.to_tornado_future(asyncio_future)

    .. versionadded:: 4.1
    """
    # Lists and dicts containing YieldPoints were handled earlier.
    if yielded is None:
        return moment
    elif isinstance(yielded, (list, dict)):
        return multi(yielded)
    elif is_future(yielded):
        return yielded
    elif isawaitable(yielded):
        return _wrap_awaitable(yielded)
    else:
        raise BadYieldError("yielded unknown object %r" % (yielded,))


if singledispatch is not None:
    convert_yielded = singledispatch(convert_yielded)

    try:
        # If we can import t.p.asyncio, do it for its side effect
        # (registering asyncio.Future with convert_yielded).
        # It's ugly to do this here, but it prevents a cryptic
        # infinite recursion in _wrap_awaitable.
        # Note that even with this, asyncio integration is unlikely
        # to work unless the application also configures AsyncIOLoop,
        # but at least the error messages in that case are more
        # comprehensible than a stack overflow.
        import salt.ext.tornado.platform.asyncio
    except ImportError:
        pass
    else:
        # Reference the imported module to make pyflakes happy.
        tornado



"""
Managing implicit state and baselines using snapshots
=====================================================

.. versionadded:: 2016.11.0

Salt can manage state against explicitly defined state, for example
if your minion state is defined by:

.. code-block:: yaml

   /etc/config_file:
     file.managed:
       - source: salt://configs/myconfig

If someone modifies this file, the next application of the highstate will
allow the admin to correct this deviation and the file will be corrected.

Now, what happens if somebody creates a file ``/etc/new_config_file`` and
deletes ``/etc/important_config_file``? Unless you have a explicit rule, this
change will go unnoticed.

The snapper state module allows you to manage state implicitly, in addition
to explicit rules, in order to define a baseline and iterate with explicit
rules as they show that they work in production.

The workflow is: once you have a working and audited system, you would create
your baseline snapshot (eg. with ``salt tgt snapper.create_snapshot``) and
define in your state this baseline using the identifier of the snapshot
(in this case: 20):

.. code-block:: yaml

    my_baseline:
      snapper.baseline_snapshot:
        - number: 20
        - include_diff: False
        - ignore:
          - /var/log
          - /var/cache

Baseline snapshots can be also referenced by tag. Most recent baseline snapshot
is used in case of multiple snapshots with the same tag:

    my_baseline_external_storage:
      snapper.baseline_snapshot:
        - tag: my_custom_baseline_tag
        - config: external
        - ignore:
          - /mnt/tmp_files/

If you have this state, and you haven't done changes to the system since the
snapshot, and you add a user, the state will show you the changes (including
full diffs) to ``/etc/passwd``, ``/etc/shadow``, etc if you call it
with ``test=True`` and will undo all changes if you call it without.

This allows you to add more explicit state knowing that you are starting from a
very well defined state, and that you can audit any change that is not part
of your explicit configuration.

So after you made this your state, you decided to introduce a change in your
configuration:

.. code-block:: yaml

    my_baseline:
      snapper.baseline_snapshot:
        - number: 20
        - ignore:
          - /var/log
          - /var/cache

    hosts_entry:
      file.blockreplace:
        - name: /etc/hosts
        - content: 'First line of content'
        - append_if_not_found: True


The change in ``/etc/hosts`` will be done after any other change that deviates
from the specified snapshot are reverted. This could be for example,
modifications to the ``/etc/passwd`` file or changes in the ``/etc/hosts``
that could render your the ``hosts_entry`` rule void or dangerous.

Once you take a new snapshot and you update the baseline snapshot number to
include the change in ``/etc/hosts`` the ``hosts_entry`` rule will basically
do nothing. You are free to leave it there for documentation, to ensure that
the change is made in case the snapshot is wrong, but if you remove anything
that comes after the ``snapper.baseline_snapshot`` as it will have no effect;
by the moment the state is evaluated, the baseline state was already applied
and include this change.

.. warning::
    Make sure you specify the baseline state before other rules, otherwise
    the baseline state will revert all changes if they are not present in
    the snapshot.

.. warning::
    Do not specify more than one baseline rule as only the last one will
    affect the result.

:codeauthor:    Duncan Mac-Vicar P. <dmacvicar@suse.de>
:codeauthor:    Pablo Suárez Hernández <psuarezhernandez@suse.de>

:maturity:      new
:platform:      Linux
"""


import os


def __virtual__():
    """
    Only load if the snapper module is available in __salt__
    """
    if "snapper.diff" in __salt__:
        return "snapper"
    return (False, "snapper module could not be loaded")


def _get_baseline_from_tag(config, tag):
    """
    Returns the last created baseline snapshot marked with `tag`
    """
    last_snapshot = None
    for snapshot in __salt__["snapper.list_snapshots"](config):
        if tag == snapshot["userdata"].get("baseline_tag"):
            if not last_snapshot or last_snapshot["timestamp"] < snapshot["timestamp"]:
                last_snapshot = snapshot
    return last_snapshot


def baseline_snapshot(
    name, number=None, tag=None, include_diff=True, config="root", ignore=None
):
    """
    Enforces that no file is modified comparing against a previously
    defined snapshot identified by number.

    number
        Number of selected baseline snapshot.

    tag
        Tag of the selected baseline snapshot. Most recent baseline baseline
        snapshot is used in case of multiple snapshots with the same tag.
        (`tag` and `number` cannot be used at the same time)

    include_diff
        Include a diff in the response (Default: True)

    config
        Snapper config name (Default: root)

    ignore
        List of files to ignore. (Default: None)
    """
    if not ignore:
        ignore = []

    ret = {"changes": {}, "comment": "", "name": name, "result": True}

    if number is None and tag is None:
        ret.update(
            {"result": False, "comment": "Snapshot tag or number must be specified"}
        )
        return ret

    if number and tag:
        ret.update(
            {
                "result": False,
                "comment": "Cannot use snapshot tag and number at the same time",
            }
        )
        return ret

    if tag:
        snapshot = _get_baseline_from_tag(config, tag)
        if not snapshot:
            ret.update(
                {"result": False, "comment": 'Baseline tag "{}" not found'.format(tag)}
            )
            return ret
        number = snapshot["id"]

    status = __salt__["snapper.status"](config, num_pre=0, num_post=number)

    for target in ignore:
        if os.path.isfile(target):
            status.pop(target, None)
        elif os.path.isdir(target):
            for target_file in [
                target_file
                for target_file in status.keys()
                if target_file.startswith(target)
            ]:
                status.pop(target_file, None)

    for file in status:
        # Only include diff for modified files
        if "modified" in status[file]["status"] and include_diff:
            status[file].pop("status")
            status[file].update(
                __salt__["snapper.diff"](
                    config, num_pre=0, num_post=number, filename=file
                ).get(file, {})
            )

    if __opts__["test"] and status:
        ret["changes"] = status
        ret["comment"] = "{} files changes are set to be undone".format(
            len(status.keys())
        )
        ret["result"] = None
    elif __opts__["test"] and not status:
        ret["changes"] = {}
        ret["comment"] = "Nothing to be done"
        ret["result"] = True
    elif not __opts__["test"] and status:
        undo = __salt__["snapper.undo"](
            config, num_pre=number, num_post=0, files=status.keys()
        )
        ret["changes"]["sumary"] = undo
        ret["changes"]["files"] = status
        ret["result"] = True
    else:
        ret["comment"] = "No changes were done"
        ret["result"] = True

    return ret



"""
    :codeauthor: :email:`Anthony Shaw <anthonyshaw@apache.org>`
"""


import salt.modules.napalm_network as napalm_network
import salt.modules.napalm_users as napalm_users
import tests.support.napalm as napalm_test_support
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class NapalmUsersModuleTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        patcher = patch(
            "salt.utils.napalm.get_device",
            MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
        )
        patcher.start()
        self.addCleanup(patcher.stop)
        module_globals = {
            "__salt__": {
                "config.get": MagicMock(
                    return_value={"test": {"driver": "test", "key": "2orgk34kgk34g"}}
                ),
                "file.file_exists": napalm_test_support.true,
                "file.join": napalm_test_support.join,
                "file.get_managed": napalm_test_support.get_managed_file,
                "random.hash": napalm_test_support.random_hash,
                "net.load_template": napalm_network.load_template,
            }
        }

        return {napalm_users: module_globals, napalm_network: module_globals}

    def test_config(self):
        ret = napalm_users.config()
        assert ret["out"] == napalm_test_support.TEST_USERS.copy()

    def test_set_users(self):
        ret = napalm_users.set_users({"mircea": {}})
        assert ret["result"] is False

    def test_delete_users(self):
        ret = napalm_users.delete_users({"mircea": {}})
        assert ret["result"] is False



"""
    :codeauthor: :email:`Anthony Shaw <anthonyshaw@apache.org>`
"""


import salt.grains.napalm as napalm_grains
import salt.proxy.napalm as napalm_proxy
import tests.support.napalm as napalm_test_support
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class NapalmGrainsTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        patcher = patch.object(
            napalm_grains.salt.utils.napalm, "is_proxy", MagicMock(return_value=True)
        )
        patcher.start()
        test_device_cache = {
            "DRIVER": napalm_test_support.MockNapalmDevice(),
            "DRIVER_NAME": "cisco",
            "OS_VERSION": "1.2.3",
            "HOSTNAME": "test-device.com",
            "USERNAME": "admin",
        }

        test_cache = {"result": True, "out": napalm_test_support.TEST_FACTS}
        patcher = patch("salt.grains.napalm.DEVICE_CACHE", test_device_cache)
        patcher.start()
        patcher = patch("salt.grains.napalm.GRAINS_CACHE", test_cache)
        patcher.start()
        self.addCleanup(patch.stopall)
        module_globals = {
            "__salt__": {
                "config.option": MagicMock(
                    return_value={"test": {"driver": "test", "key": "2orgk34kgk34g"}}
                ),
                "file.file_exists": napalm_test_support.true,
                "file.join": napalm_test_support.join,
                "file.get_managed": napalm_test_support.get_managed_file,
                "random.hash": napalm_test_support.random_hash,
            }
        }

        return {napalm_grains: module_globals}

    def test_os(self):
        ret = napalm_grains.getos(proxy=napalm_proxy)
        assert ret["os"] == "cisco"

    def test_os_version(self):
        ret = napalm_grains.version(proxy=napalm_proxy)
        assert ret["version"] == "1.2.3"

    def test_model(self):
        ret = napalm_grains.model(proxy=napalm_proxy)
        assert ret["model"] == "test_model"

    def test_serial(self):
        ret = napalm_grains.serial(proxy=napalm_proxy)
        assert ret["serial"] == "123456"

    def test_vendor(self):
        ret = napalm_grains.vendor(proxy=napalm_proxy)
        assert ret["vendor"] == "cisco"

    def test_uptime(self):
        ret = napalm_grains.uptime(proxy=napalm_proxy)
        assert ret["uptime"] == "Forever"

    def test_interfaces(self):
        ret = napalm_grains.interfaces(proxy=napalm_proxy)
        assert ret["interfaces"] == napalm_test_support.TEST_INTERFACES

    def test_username(self):
        ret = napalm_grains.username(proxy=napalm_proxy)
        assert ret["username"] == "admin"

    def test_hostname(self):
        ret = napalm_grains.hostname(proxy=napalm_proxy)
        assert ret["hostname"] == "test-device.com"

    def test_host(self):
        ret = napalm_grains.host(proxy=napalm_proxy)
        assert ret["host"] == "test-device.com"



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""


import salt.modules.dig as dig
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

_SPF_VALUES = {
    "dig +short xmission.com TXT": {
        "pid": 27282,
        "retcode": 0,
        "stderr": "",
        "stdout": '"v=spf1 a mx include:_spf.xmission.com ?all"',
    },
    "dig +short _spf.xmission.com TXT": {
        "pid": 27282,
        "retcode": 0,
        "stderr": "",
        "stdout": '"v=spf1 a mx ip4:198.60.22.0/24 ip4:166.70.13.0/24 ~all"',
    },
    "dig +short xmission-redirect.com TXT": {
        "pid": 27282,
        "retcode": 0,
        "stderr": "",
        "stdout": "v=spf1 redirect=_spf.xmission.com",
    },
    "dig +short foo.com TXT": {
        "pid": 27282,
        "retcode": 0,
        "stderr": "",
        "stdout": "v=spf1 ip4:216.73.93.70/31 ip4:216.73.93.72/31 ~all",
    },
}


def _spf_side_effect(key, python_shell=False):
    return _SPF_VALUES.get(
        " ".join(key), {"pid": 27310, "retcode": 0, "stderr": "", "stdout": ""}
    )


@skipIf(dig.__virtual__() is False, "Dig must be installed")
class DigTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {dig: {}}

    def test_check_ip(self):
        self.assertTrue(dig.check_ip("127.0.0.1"), msg="Not a valid ip address")

    def test_check_ip_ipv6(self):
        self.assertTrue(
            dig.check_ip("1111:2222:3333:4444:5555:6666:7777:8888"),
            msg="Not a valid ip address",
        )

    def test_check_ip_ipv6_valid(self):
        self.assertTrue(dig.check_ip("2607:fa18:0:3::4"))

    def test_check_ip_neg(self):
        self.assertFalse(
            dig.check_ip("-127.0.0.1"), msg="Did not detect negative value as invalid"
        )

    def test_check_ip_empty(self):
        self.assertFalse(dig.check_ip(""), msg="Did not detect empty value as invalid")

    def test_a(self):
        dig_mock = MagicMock(
            return_value={
                "pid": 3656,
                "retcode": 0,
                "stderr": "",
                "stdout": (
                    "74.125.193.104\n"
                    "74.125.193.105\n"
                    "74.125.193.99\n"
                    "74.125.193.106\n"
                    "74.125.193.103\n"
                    "74.125.193.147"
                ),
            }
        )
        with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
            self.assertEqual(
                dig.A("www.google.com"),
                [
                    "74.125.193.104",
                    "74.125.193.105",
                    "74.125.193.99",
                    "74.125.193.106",
                    "74.125.193.103",
                    "74.125.193.147",
                ],
            )

    def test_aaaa(self):
        dig_mock = MagicMock(
            return_value={
                "pid": 25451,
                "retcode": 0,
                "stderr": "",
                "stdout": "2607:f8b0:400f:801::1014",
            }
        )
        with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
            self.assertEqual(dig.AAAA("www.google.com"), ["2607:f8b0:400f:801::1014"])

    def test_ns(self):
        with patch("salt.modules.dig.A", MagicMock(return_value=["ns4.google.com."])):
            dig_mock = MagicMock(
                return_value={
                    "pid": 26136,
                    "retcode": 0,
                    "stderr": "",
                    "stdout": "ns4.google.com.",
                }
            )
            with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
                self.assertEqual(dig.NS("google.com"), ["ns4.google.com."])

    def test_spf(self):
        dig_mock = MagicMock(side_effect=_spf_side_effect)
        with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
            self.assertEqual(dig.SPF("foo.com"), ["216.73.93.70/31", "216.73.93.72/31"])

    def test_spf_redir(self):
        """
        Test for SPF records which use the 'redirect' SPF mechanism
        https://en.wikipedia.org/wiki/Sender_Policy_Framework#Mechanisms
        """
        dig_mock = MagicMock(side_effect=_spf_side_effect)
        with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
            self.assertEqual(
                dig.SPF("xmission-redirect.com"), ["198.60.22.0/24", "166.70.13.0/24"]
            )

    def test_spf_include(self):
        """
        Test for SPF records which use the 'include' SPF mechanism
        https://en.wikipedia.org/wiki/Sender_Policy_Framework#Mechanisms
        """
        dig_mock = MagicMock(side_effect=_spf_side_effect)
        with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
            self.assertEqual(
                dig.SPF("xmission.com"), ["198.60.22.0/24", "166.70.13.0/24"]
            )

    def test_mx(self):
        dig_mock = MagicMock(
            return_value={
                "pid": 27780,
                "retcode": 0,
                "stderr": "",
                "stdout": (
                    "10 aspmx.l.google.com.\n"
                    "20 alt1.aspmx.l.google.com.\n"
                    "40 alt3.aspmx.l.google.com.\n"
                    "50 alt4.aspmx.l.google.com.\n"
                    "30 alt2.aspmx.l.google.com."
                ),
            }
        )
        with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
            self.assertEqual(
                dig.MX("google.com"),
                [
                    ["10", "aspmx.l.google.com."],
                    ["20", "alt1.aspmx.l.google.com."],
                    ["40", "alt3.aspmx.l.google.com."],
                    ["50", "alt4.aspmx.l.google.com."],
                    ["30", "alt2.aspmx.l.google.com."],
                ],
            )



"""
Wheel system wrapper for connected minions
"""


import salt.config
import salt.utils.minions
from salt.utils.cache import CacheCli


def connected():
    """
    List all connected minions on a salt-master
    """
    opts = salt.config.master_config(__opts__["conf_file"])

    if opts.get("con_cache"):
        cache_cli = CacheCli(opts)
        minions = cache_cli.get_cached()
    else:
        minions = list(salt.utils.minions.CkMinions(opts).connected_ids())
    return minions



"""
    :codeauthor: :email:`Jorge Schrauwen <sjorge@blackdot.be>`
"""

import salt.grains.mdata as mdata
from tests.support.mock import Mock, patch
from tests.support.unit import TestCase


class MdataGrainsTestCase(TestCase):
    """
    Test cases for mdata grains
    """

    def setup_loader_modules(self):
        return {
            mdata: {"__salt__": {}},
        }

    def test_user_mdata_missing_cmd_both(self):
        """
        When both or either of the commands is missing there should
        be no grain output.
        """
        grains_exp_res = {}

        which_mock = Mock(side_effect=[None, None])
        with patch("salt.utils.path.which", which_mock):
            grains_res = mdata._user_mdata()
            self.assertEqual(grains_exp_res, grains_res)

    def test_user_mdata_missing_cmd_one(self):
        """
        When both or either of the commands is missing there should
        be no grain output.
        """
        grains_exp_res = {}

        which_mock = Mock(side_effect=["/usr/sbin/mdata-list", None])
        with patch("salt.utils.path.which", which_mock):
            grains_res = mdata._user_mdata()
            self.assertEqual(grains_exp_res, grains_res)

    def test_user_mdata_empty_list(self):
        """
        When there are no user grains, there are no mdata-get calls
        so there are also no grains.
        """
        grains_exp_res = {}

        which_mock = Mock(side_effect=["/usr/sbin/mdata-list", "/usr/sbin/mdata-get"])
        cmd_mock = Mock(side_effect=[""])
        with patch("salt.utils.path.which", which_mock), patch.dict(
            mdata.__salt__, {"cmd.run": cmd_mock}
        ):
            grains_res = mdata._user_mdata()
            self.assertEqual(grains_exp_res, grains_res)

    def test_user_mdata(self):
        """
        We have a list of two grains, so there should be two mdata-get
        calls, resulting in 2 grains.
        """
        grains_exp_res = {
            "mdata": {
                "multi_text_data": "multi\nline\ntext",
                "simple_text_data": "some text data",
            },
        }

        which_mock = Mock(side_effect=["/usr/sbin/mdata-list", "/usr/sbin/mdata-get"])
        cmd_mock = Mock(
            side_effect=[
                "simple_text_data\nmulti_text_data",
                "some text data",
                "multi\nline\ntext",
            ]
        )
        with patch("salt.utils.path.which", which_mock), patch.dict(
            mdata.__salt__, {"cmd.run": cmd_mock}
        ):
            grains_res = mdata._user_mdata()

            self.assertEqual(grains_exp_res, grains_res)

    def test_sdc_mdata_missing_cmd_both(self):
        """
        When both or either of the commands is missing there should
        be no grain output.
        """
        which_mock = Mock(side_effect=[None, None])
        with patch("salt.utils.path.which", which_mock):
            grains = mdata._sdc_mdata()
            assert grains == {}

    def test_sdc_mdata_missing_cmd_one(self):
        """
        When both or either of the commands is missing there should
        be no grain output.
        """
        grains_exp_res = {}

        which_mock = Mock(side_effect=["/usr/sbin/mdata-list", None])
        with patch("salt.utils.path.which", which_mock):
            grains_res = mdata._sdc_mdata()
            self.assertEqual(grains_exp_res, grains_res)

    def test_sdc_mdata(self):
        """
        Simulate all mdata_get calls from a test zone.
        """
        grains_exp_res = {
            "mdata": {
                "sdc": {
                    "alias": "test",
                    "dns_domain": "example.org",
                    "hostname": "test_salt",
                    "nics": [
                        {
                            "gateway": "10.12.3.1",
                            "gateways": ["10.12.3.1"],
                            "interface": "net0",
                            "ip": "10.12.3.123",
                            "ips": ["10.12.3.123/24", "2001:ffff:ffff:123::123/64"],
                            "mac": "00:00:00:00:00:01",
                            "mtu": 1500,
                            "netmask": "255.255.255.0",
                            "nic_tag": "trunk",
                            "primary": True,
                            "vlan_id": 123,
                        }
                    ],
                    "resolvers": ["10.12.3.1", "2001:ffff:ffff:123::1"],
                    "routes": [],
                    "server_uuid": "00000000-0000-0000-0000-000123456789",
                    "uuid": "bae504b1-4594-47de-e2ed-e4f454776689",
                },
            },
        }

        which_mock = Mock(side_effect=["/usr/sbin/mdata-list", "/usr/sbin/mdata-get"])
        cmd_mock = Mock(
            side_effect=[
                "bae504b1-4594-47de-e2ed-e4f454776689",
                "00000000-0000-0000-0000-000123456789",
                "No metadata for 'sdc:datacenter_name'",
                "test_salt",
                "example.org",
                "test",
                '["10.12.3.1","2001:ffff:ffff:123::1"]',
                '[{"interface":"net0","mac":"00:00:00:00:00:01","vlan_id":123,"nic_tag":"trunk","gateway":"10.12.3.1","gateways":["10.12.3.1"],"netmask":"255.255.255.0","ip":"10.12.3.123","ips":["10.12.3.123/24","2001:ffff:ffff:123::123/64"],"mtu":1500,"primary":true}]',
                "[]",
            ]
        )
        with patch("salt.utils.path.which", which_mock), patch.dict(
            mdata.__salt__, {"cmd.run": cmd_mock}
        ):
            grains_res = mdata._sdc_mdata()

            self.assertEqual(grains_exp_res, grains_res)



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""


import logging

import salt.modules.dnsutil as dnsutil
import salt.utils.stringutils
from tests.support.mock import MagicMock, mock_open, patch
from tests.support.unit import TestCase, skipIf

log = logging.getLogger(__name__)

mock_hosts_file = salt.utils.stringutils.to_str(
    "##\n"
    "# Host Database\n"
    "#\n"
    "# localhost is used to configure the loopback interface\n"
    "# when the system is booting.  Do not change this entry.\n"
    "##\n"
    "127.0.0.1	localhost\n"
    "255.255.255.255	broadcasthost\n"
    "::1             localhost\n"
    "fe80::1%lo0	localhost"
)

mock_hosts_file_rtn = {
    "::1": ["localhost"],
    "255.255.255.255": ["broadcasthost"],
    "127.0.0.1": ["localhost"],
    "fe80::1%lo0": ["localhost"],
}

mock_soa_zone = salt.utils.stringutils.to_str(
    "$TTL 3D\n"
    "@               IN      SOA     land-5.com. root.land-5.com. (\n"
    "199609203       ; Serial\n"
    "28800   ; Refresh\n"
    "7200    ; Retry\n"
    "604800  ; Expire\n"
    "86400)  ; Minimum TTL\n"
    "NS      land-5.com.\n\n"
    "1                       PTR     localhost."
)

mock_writes_list = salt.utils.data.decode(
    [
        "##\n",
        "# Host Database\n",
        "#\n",
        "# localhost is used to configure the loopback interface\n",
        "# when the system is booting.  Do not change this entry.\n",
        "##\n",
        "127.0.0.1 localhost",
        "\n",
        "255.255.255.255 broadcasthost",
        "\n",
        "::1 localhost",
        "\n",
        "fe80::1%lo0 localhost",
        "\n",
    ],
    to_str=True,
)


class DNSUtilTestCase(TestCase):
    def test_parse_hosts(self):
        with patch("salt.utils.files.fopen", mock_open(read_data=mock_hosts_file)):
            self.assertEqual(
                dnsutil.parse_hosts(),
                {
                    "::1": ["localhost"],
                    "255.255.255.255": ["broadcasthost"],
                    "127.0.0.1": ["localhost"],
                    "fe80::1%lo0": ["localhost"],
                },
            )

    def test_hosts_append(self):
        with patch(
            "salt.utils.files.fopen", mock_open(read_data=mock_hosts_file)
        ) as m_open, patch(
            "salt.modules.dnsutil.parse_hosts",
            MagicMock(return_value=mock_hosts_file_rtn),
        ):
            dnsutil.hosts_append("/etc/hosts", "127.0.0.1", "ad1.yuk.co,ad2.yuk.co")
            writes = m_open.write_calls()
            # We should have called .write() only once, with the expected
            # content
            num_writes = len(writes)
            assert num_writes == 1, num_writes
            expected = salt.utils.stringutils.to_str(
                "\n127.0.0.1 ad1.yuk.co ad2.yuk.co"
            )
            assert writes[0] == expected, writes[0]

    def test_hosts_remove(self):
        to_remove = "ad1.yuk.co"
        new_mock_file = mock_hosts_file + "\n127.0.0.1 " + to_remove + "\n"
        with patch(
            "salt.utils.files.fopen", mock_open(read_data=new_mock_file)
        ) as m_open:
            dnsutil.hosts_remove("/etc/hosts", to_remove)
            writes = m_open.write_calls()
            assert writes == mock_writes_list, writes

    @skipIf(True, "Waiting on bug report fixes")
    def test_parse_zone(self):
        with patch("salt.utils.files.fopen", mock_open(read_data=mock_soa_zone)):
            log.debug(mock_soa_zone)
            log.debug(dnsutil.parse_zone("/var/lib/named/example.com.zone"))

    def test_to_seconds_hour(self):
        self.assertEqual(
            dnsutil._to_seconds("4H"),
            14400,
            msg="Did not detect valid hours as invalid",
        )

    def test_to_seconds_day(self):
        self.assertEqual(
            dnsutil._to_seconds("1D"), 86400, msg="Did not detect valid day as invalid"
        )

    def test_to_seconds_week(self):
        self.assertEqual(
            dnsutil._to_seconds("2W"),
            604800,
            msg="Did not set time greater than one week to one week",
        )

    def test_to_seconds_empty(self):
        self.assertEqual(
            dnsutil._to_seconds(""), 604800, msg="Did not set empty time to one week"
        )

    def test_to_seconds_large(self):
        self.assertEqual(
            dnsutil._to_seconds("604801"),
            604800,
            msg="Did not set time greater than one week to one week",
        )



"""
uWSGI stats server https://uwsgi-docs.readthedocs.io/en/latest/StatsServer.html

:maintainer: Peter Baumgartner <pete@lincolnloop.com>
:maturity:   new
:platform:   all
"""

import salt.utils.json
import salt.utils.path


def __virtual__():
    """
    Only load the module if uwsgi is installed
    """
    cmd = "uwsgi"
    if salt.utils.path.which(cmd):
        return cmd
    return (
        False,
        "The uwsgi execution module failed to load: the uwsgi binary is not in the"
        " path.",
    )


def stats(socket):
    """
    Return the data from `uwsgi --connect-and-read` as a dictionary.

    socket
        The socket the uWSGI stats server is listening on

    CLI Example:

    .. code-block:: bash

        salt '*' uwsgi.stats /var/run/mystatsserver.sock

        salt '*' uwsgi.stats 127.0.0.1:5050
    """

    cmd = ["uwsgi", "--connect-and-read", "{}".format(socket)]
    out = __salt__["cmd.run"](cmd, python_shell=False)
    return salt.utils.json.loads(out)



"""
States for managing zpools

:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:depends:       salt.utils.zfs, salt.modules.zpool
:platform:      smartos, illumos, solaris, freebsd, linux

.. versionadded:: 2016.3.0
.. versionchanged:: 2018.3.1
  Big refactor to remove duplicate code, better type conversions and improved
  consistency in output.

.. code-block:: yaml

    oldpool:
      zpool.absent:
        - export: true

    newpool:
      zpool.present:
        - config:
            import: false
            force: true
        - properties:
            comment: salty storage pool
        - layout:
            - mirror:
              - /dev/disk0
              - /dev/disk1
            - mirror:
              - /dev/disk2
              - /dev/disk3

    partitionpool:
      zpool.present:
        - config:
            import: false
            force: true
        - properties:
            comment: disk partition salty storage pool
            ashift: '12'
            feature@lz4_compress: enabled
        - filesystem_properties:
            compression: lz4
            atime: on
            relatime: on
        - layout:
            - /dev/disk/by-uuid/3e43ce94-77af-4f52-a91b-6cdbb0b0f41b

    simplepool:
      zpool.present:
        - config:
            import: false
            force: true
        - properties:
            comment: another salty storage pool
        - layout:
            - /dev/disk0
            - /dev/disk1

.. warning::

    The layout will never be updated, it will only be used at time of creation.
    It's a whole lot of work to figure out if a devices needs to be detached, removed,
    etc. This is best done by the sysadmin on a case per case basis.

    Filesystem properties are also not updated, this should be managed by the zfs state module.

"""

import logging
import os

from salt.utils.odict import OrderedDict

log = logging.getLogger(__name__)

# Define the state's virtual name
__virtualname__ = "zpool"


def __virtual__():
    """
    Provides zpool state
    """
    if not __grains__.get("zfs_support"):
        return False, "The zpool state cannot be loaded: zfs not supported"
    return __virtualname__


def _layout_to_vdev(layout, device_dir=None):
    """
    Turn the layout data into usable vdevs spedcification

    We need to support 2 ways of passing the layout:

    .. code::
        layout_new:
          - mirror:
            - disk0
            - disk1
          - mirror:
            - disk2
            - disk3

    .. code:
        layout_legacy:
          mirror-0:
            disk0
            disk1
          mirror-1:
            disk2
            disk3

    """
    vdevs = []

    # NOTE: check device_dir exists
    if device_dir and not os.path.exists(device_dir):
        device_dir = None

    # NOTE: handle list of OrderedDicts (new layout)
    if isinstance(layout, list):
        # NOTE: parse each vdev as a tiny layout and just append
        for vdev in layout:
            if isinstance(vdev, OrderedDict):
                vdevs.extend(_layout_to_vdev(vdev, device_dir))
            else:
                if device_dir and vdev[0] != "/":
                    vdev = os.path.join(device_dir, vdev)
                vdevs.append(vdev)

    # NOTE: handle nested OrderedDict (legacy layout)
    #       this is also used to parse the nested OrderedDicts
    #       from the new layout
    elif isinstance(layout, OrderedDict):
        for vdev in layout:
            # NOTE: extract the vdev type and disks in the vdev
            vdev_type = vdev.split("-")[0]
            vdev_disk = layout[vdev]

            # NOTE: skip appending the dummy type 'disk'
            if vdev_type != "disk":
                vdevs.append(vdev_type)

            # NOTE: ensure the disks are a list (legacy layout are not)
            if not isinstance(vdev_disk, list):
                vdev_disk = vdev_disk.split(" ")

            # NOTE: also append the actualy disks behind the type
            #       also prepend device_dir to disks if required
            for disk in vdev_disk:
                if device_dir and disk[0] != "/":
                    disk = os.path.join(device_dir, disk)
                vdevs.append(disk)

    # NOTE: we got invalid data for layout
    else:
        vdevs = None

    return vdevs


def present(
    name, properties=None, filesystem_properties=None, layout=None, config=None
):
    """
    ensure storage pool is present on the system

    name : string
        name of storage pool
    properties : dict
        optional set of properties to set for the storage pool
    filesystem_properties : dict
        optional set of filesystem properties to set for the storage pool (creation only)
    layout: dict
        disk layout to use if the pool does not exist (creation only)
    config : dict
        fine grain control over this state

    .. note::

        The following configuration properties can be toggled in the config parameter.
          - import (true) - try to import the pool before creating it if absent
          - import_dirs (None) - specify additional locations to scan for devices on import (comma-separated)
          - device_dir (None, SunOS=/dev/dsk, Linux=/dev) - specify device directory to prepend for none
            absolute device paths
          - force (false) - try to force the import or creation

    .. note::

        It is no longer needed to give a unique name to each top-level vdev, the old
        layout format is still supported but no longer recommended.

        .. code-block:: yaml

            - mirror:
              - /tmp/vdisk3
              - /tmp/vdisk2
            - mirror:
              - /tmp/vdisk0
              - /tmp/vdisk1

        The above yaml will always result in the following zpool create:

        .. code-block:: bash

            zpool create mypool mirror /tmp/vdisk3 /tmp/vdisk2 mirror /tmp/vdisk0 /tmp/vdisk1

    .. warning::

        The legacy format is also still supported but not recommended,
        because ID's inside the layout dict must be unique they need to have a suffix.

        .. code-block:: yaml

            mirror-0:
              /tmp/vdisk3
              /tmp/vdisk2
            mirror-1:
              /tmp/vdisk0
              /tmp/vdisk1

    .. warning::

        Pay attention to the order of your dict!

        .. code-block:: yaml

            - mirror:
              - /tmp/vdisk0
              - /tmp/vdisk1
            - /tmp/vdisk2

        The above will result in the following zpool create:

        .. code-block:: bash

            zpool create mypool mirror /tmp/vdisk0 /tmp/vdisk1 /tmp/vdisk2

        Creating a 3-way mirror! While you probably expect it to be mirror
        root vdev with 2 devices + a root vdev of 1 device!

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    # config defaults
    default_config = {
        "import": True,
        "import_dirs": None,
        "device_dir": None,
        "force": False,
    }
    if __grains__["kernel"] == "SunOS":
        default_config["device_dir"] = "/dev/dsk"
    elif __grains__["kernel"] == "Linux":
        default_config["device_dir"] = "/dev"

    # merge state config
    if config:
        default_config.update(config)
    config = default_config

    # ensure properties are zfs values
    if properties:
        properties = __utils__["zfs.from_auto_dict"](properties)
    elif properties is None:
        properties = {}
    if filesystem_properties:
        filesystem_properties = __utils__["zfs.from_auto_dict"](filesystem_properties)
    elif filesystem_properties is None:
        filesystem_properties = {}

    # parse layout
    vdevs = _layout_to_vdev(layout, config["device_dir"])
    if vdevs:
        vdevs.insert(0, name)

    # log configuration
    log.debug("zpool.present::%s::config - %s", name, config)
    log.debug("zpool.present::%s::vdevs - %s", name, vdevs)
    log.debug("zpool.present::%s::properties -  %s", name, properties)
    log.debug(
        "zpool.present::%s::filesystem_properties -  %s", name, filesystem_properties
    )

    # ensure the pool is present
    ret["result"] = False

    # don't do anything because this is a test
    if __opts__["test"]:
        if __salt__["zpool.exists"](name):
            ret["result"] = True
            ret["comment"] = "storage pool {} is {}".format(name, "uptodate")
        else:
            ret["result"] = None
            ret["changes"][name] = "imported" if config["import"] else "created"
            ret["comment"] = "storage pool {} would have been {}".format(
                name, ret["changes"][name]
            )

    # update pool
    elif __salt__["zpool.exists"](name):
        ret["result"] = True

        # fetch current pool properties
        properties_current = __salt__["zpool.get"](name, parsable=True)

        # build list of properties to update
        properties_update = []
        if properties:
            for prop in properties:
                # skip unexisting properties
                if prop not in properties_current:
                    log.warning(
                        "zpool.present::%s::update - unknown property: %s", name, prop
                    )
                    continue

                # compare current and wanted value
                if properties_current[prop] != properties[prop]:
                    properties_update.append(prop)

        # update pool properties
        for prop in properties_update:
            res = __salt__["zpool.set"](name, prop, properties[prop])

            if res["set"]:
                if name not in ret["changes"]:
                    ret["changes"][name] = {}
                ret["changes"][name][prop] = properties[prop]
            else:
                ret["result"] = False
                if ret["comment"] == "":
                    ret["comment"] = "The following properties were not updated:"
                ret["comment"] = "{} {}".format(ret["comment"], prop)

        if ret["result"]:
            ret["comment"] = (
                "properties updated" if ret["changes"] else "no update needed"
            )

    # import or create the pool (at least try to anyway)
    else:
        # import pool
        if config["import"]:
            mod_res = __salt__["zpool.import"](
                name,
                force=config["force"],
                dir=config["import_dirs"],
            )

            ret["result"] = mod_res["imported"]
            if ret["result"]:
                ret["changes"][name] = "imported"
                ret["comment"] = "storage pool {} was imported".format(name)

        # create pool
        if not ret["result"] and vdevs:
            log.debug("zpool.present::%s::creating", name)

            # execute zpool.create
            mod_res = __salt__["zpool.create"](
                *vdevs,
                force=config["force"],
                properties=properties,
                filesystem_properties=filesystem_properties
            )

            ret["result"] = mod_res["created"]
            if ret["result"]:
                ret["changes"][name] = "created"
                ret["comment"] = "storage pool {} was created".format(name)
            elif "error" in mod_res:
                ret["comment"] = mod_res["error"]
            else:
                ret["comment"] = "could not create storage pool {}".format(name)

        # give up, we cannot import the pool and we do not have a layout to create it
        if not ret["result"] and not vdevs:
            ret["comment"] = (
                "storage pool {} was not imported, no (valid) layout specified for"
                " creation".format(name)
            )

    return ret


def absent(name, export=False, force=False):
    """
    ensure storage pool is absent on the system

    name : string
        name of storage pool
    export : boolean
        export instead of destroy the zpool if present
    force : boolean
        force destroy or export

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    # log configuration
    log.debug("zpool.absent::%s::config::force = %s", name, force)
    log.debug("zpool.absent::%s::config::export = %s", name, export)

    # ensure the pool is absent
    if __salt__["zpool.exists"](name):  # looks like we need to do some work
        mod_res = {}
        ret["result"] = False

        # NOTE: handle test
        if __opts__["test"]:
            ret["result"] = True

        # NOTE: try to export the pool
        elif export:
            mod_res = __salt__["zpool.export"](name, force=force)
            ret["result"] = mod_res["exported"]

        # NOTE: try to destroy the pool
        else:
            mod_res = __salt__["zpool.destroy"](name, force=force)
            ret["result"] = mod_res["destroyed"]

        if ret["result"]:  # update the changes and comment
            ret["changes"][name] = "exported" if export else "destroyed"
            ret["comment"] = "storage pool {} was {}".format(name, ret["changes"][name])
        elif "error" in mod_res:
            ret["comment"] = mod_res["error"]

    else:  # we are looking good
        ret["result"] = True
        ret["comment"] = "storage pool {} is absent".format(name)

    return ret


# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
